from __future__ import annotations
from itertools import chain, repeat
from pathlib import Path
import math

from .core import *
from .machines import *
from . import itm, rcp, rcpinst, config
from .data import entityToMachine

__all__ = ('Blueprint', 'BlueprintBook', 'decodeBlueprint')

class Blueprint:
    def __init__(self, bp):
        """Create a blueprint from a decoded JSON object.

        See also `decodeBlueprint`.
        """
        if 'blueprint' not in bp:
            raise ValueError('expected a blueprint')
        self.raw = bp
    def convert(self, *, burnerFuel=None, rocketSiloRecipe='space-science-pack') -> Group:
        """Convert a blueprint into a `Group`."""
        recipeMap = rcpinst.byName[config.mode.get()]
        if burnerFuel is None:
            burnerFuel = config.defaultFuel.get()
        
        machines = []
        beacons = []
        beaconsOnGrid = {}

        for v in self.raw['blueprint']['entities']:
            try:
                cls = entityToMachine[v['name']]
            except KeyError:
                continue
            m = cls()
            m.blueprintInfo = v
            if type(m) is RocketSilo:
                m.recipe = recipeMap[rocketSiloRecipe]
            elif 'recipe' in v:
                m.recipe = recipeMap[v['recipe']]
            if hasattr(m, 'fuel'):
                m.fuel = burnerFuel 
            if 'items' in v:
                m.modules = [*chain.from_iterable(repeat(itm.byName[item], num) for item, num in v['items'].items())]
            if cls is Beacon:
                beacons.append(m)
            else:
                machines.append(m)

        for m in beacons:
            x = math.floor(m.blueprintInfo['position']['x'])
            y = math.floor(m.blueprintInfo['position']['y'])
            beaconsOnGrid.setdefault(x, {})[y] = m

        for m in machines:
            if not hasattr(m, 'beacons'):
                continue
            def as_int(x):
                if not x.is_integer():
                    raise ValeuError
                return int(x)
            x_min = as_int(m.blueprintInfo['position']['x'] - m.width/2)
            x_max = as_int(m.blueprintInfo['position']['x'] + m.width/2)
            y_min = as_int(m.blueprintInfo['position']['y'] - m.height/2)
            y_max = as_int(m.blueprintInfo['position']['y'] + m.height/2)
            for x in range(x_min-4,x_max+4):
                yp = beaconsOnGrid.get(x, {})
                for y in range(y_min-4,y_max+4):
                    beacon = yp.get(y, None)
                    if beacon is not None:
                        m.beacons.append(beacon)
                        pass

        return Group(Group(machines), Group(beacons))

class BlueprintBook:
    def __init__(self, bp):
        """Create a blueprint book from a decoded JSON object.

        See also :py:obj:`decodeBlueprint`.
        """
        if 'blueprint_book' not in bp:
            raise ValueError('expected a blueprint book')
        self.raw = bp
    def __getitem__(self, idx):
        return Blueprint(self.raw['blueprint_book']['blueprints'][idx])
    def __len__(self):
        return len(self.raw['blueprint_book']['blueprints'])
    def find(self, label) -> Blueprint:
        """Find the blueprint with *label*."""
        return BlueprintBook._find(self.raw, label)
    @staticmethod
    def _find(bp, label):
        if 'blueprint' in bp:
            if bp['blueprint'].get('label',None) == label:
                return Blueprint(bp)
        elif 'blueprint_book' in bp:
            for b in bp['blueprint_book']['blueprints']:
                try:
                    return BlueprintBook._find(b, label)
                except KeyError:
                    pass
        raise KeyError(label)
    def labels(self) -> list[str]:
        lst = []
        BlueprintBook._labels(self.raw, lst)
        return lst
    @staticmethod
    def _labels(bp, lst):
        if 'blueprint' in bp:
            lst.append(bp['blueprint'].get('label',None))
        elif 'blueprint_book' in bp:
            for b in bp['blueprint_book']['blueprints']:
                BlueprintBook._labels(b, lst)
    def convert(self, label, **convertArgs) -> Group:
        """Convert a blueprint with *label* into a `Group`"""
        return self.find(label).convert(**convertArgs)

def decodeBlueprint(bp):
    """Decode a string or file (if `pathlib.Path <https://docs.python.org/3/library/pathlib.html>`_) to a blueprint or blueprint book.
    """
    import zlib
    from base64 import b64decode
    import json

    if isinstance(bp,Path):
        bpStr = bp.read_bytes()
    else:
        bpStr = bp

    decoded = b64decode(bpStr[1:])
    decompressed = zlib.decompress(decoded)
    json = json.loads(decompressed)

    if 'blueprint_book' in json:
        return BlueprintBook(json)
    else:
        return Blueprint(json)
    



from contextvars import ContextVar,copy_context
from copy import copy
from . import itm, machines as mch
from . import config

__all__ = ('MP_EARLY_GAME', 'MP_LATE_GAME',
           'MP_MAX_PROD', 'MachinePrefs', 'SPEED_BEACON', 'withSettings')

class MachinePrefs(tuple):
    def __new__(cls, *args):
        return tuple.__new__(cls, args)
    def __add__(self, other):
        return tuple.__new__(MachinePrefs, tuple.__add__(self, other))
    def __radd__(self, other):
        if isinstance(other, tuple):
            return tuple.__new__(MachinePrefs, tuple.__add__(other, self))
        else:
            return NotImplemented
    def __mul__(self, num):
        raise NotImplementedError
    __rmul__ = __mul__
    def withSpeedBeacons(self, mapping):
        lst = list(self)
        for i, m in enumerate(lst):
            cls = type(m)
            try:
                numBeacons = mapping[cls]
            except KeyError:
                pass
            else:
                m = copy(m)
                m.beacons = numBeacons * SPEED_BEACON
                lst[i] = m
        return MachinePrefs(*lst)

MP_EARLY_GAME = MachinePrefs(mch.AssemblingMachine1(), mch.StoneFurnance())
 
MP_LATE_GAME = MachinePrefs(mch.AssemblingMachine3(), mch.ElectricFurnace())

MP_MAX_PROD = MachinePrefs(
    mch.AssemblingMachine3(modules=4*[itm.productivity_module_3]),
    mch.ElectricFurnace(modules=2*[itm.productivity_module_3]),
    mch.ChemicalPlant(modules=3*[itm.productivity_module_3]),
    mch.OilRefinery(modules=3*[itm.productivity_module_3]),
    mch.RocketSilo(modules=4*[itm.productivity_module_3]),
    mch.Centrifuge(modules=2*[itm.productivity_module_3]),
)

SPEED_BEACON = mch.Beacon(modules=[itm.speed_module_3, itm.speed_module_3])


def withSettings(settings, fun, *args, **kwargs):
    "helper function to locally set a ContextVar"
    ctx = copy_context()
    return ctx.run(__withSettings, settings, fun, *args, **kwargs)
        

def __withSettings(settings, fun, *args, **kwargs):
    for var, val in settings.items():
        var.set(val)
    return fun(*args, **kwargs)
    



from .core import *

def toPythonName(name):
    return name.replace('-','_')

def asItem(item):
    if item is None:
        return item
    if isinstance(item, str):
        item = itm.byName[item]
    if not isinstance(item, Ingredient):
        raise TypeError(f'invalid type for item: {type(item)}')
    return item



from __future__ import annotations
from dataclasses import dataclass,field
from collections.abc import Mapping
from copy import copy as _copy
from numbers import Number
import sys

from .fracs import frac, div, ceil, Inf
from .core import *
from .core import _MutableFlows,NetFlows
from ._helper import asItem
from . import itm

__all__ = ('BoxBase', 'Box', 'UnboundedBox', 'BlackBox')

class BoxFlows(NetFlows):
    def _showFlow(self, flow):
        return flow.rateOut != flow.rateIn or flow.rateOut == 0 == flow.rateIn

@dataclass(init=False)
class BoxBase(MachineBase):
    @property
    def num(self):
        return 1

    @property
    def machine(self):
        return self

    @property
    def recipe(self):
        return None

    def __repr__(self):
        nameStr = f' ({self.name})' if self.name is not None else ''
        throttleStr = f' @{self.throttle:g}' if getattr(self, 'throttle', 1) != 1 else ''
        return f'<{self.__class__.__name__}{nameStr}{throttleStr} {self.outputs} <= {self.inputs}>'

    def __str__(self):
        name = self.name if self.name is not None else self._fallbackName
        throttleStr = f' @{self.throttle:g}' if getattr(self, 'throttle', 1) != 1 else ''
        return f'{name}{throttleStr}: {self.outputs} <= {self.inputs}'

    def _flatten(self, lst, num):
        if num == 1:
            lst.append(self)
        else:
            lst.append(Mul(self, num))

class Box(BoxBase):
    """Wrap a group to restrict inputs or outputs."""
    MIN_PRIORITY =  IGNORE # (-100)
    MAX_PRIORITY = -IGNORE #  (100)

    _fallbackName = '<box>'

    class _ExternalFlows(dict):
        __slots__ = ()
        def __init__(self, vals = None):
            self.update(vals)
        def update(self, vals = None):
            if vals is None:
                return
            for v in vals.items() if isinstance(vals, Mapping) else vals:
                if type(v) is tuple:
                    item, rate = v
                elif isinstance(v, OneWayFlow):
                    item, rate = v.item, v.rate
                elif isinstance(v, Flow):
                    item, rate = v.item, abs(v.rate())
                else:
                    item, rate = v, None
                self[item] = rate
        def add(self, item, rate = None):
            self[item] = rate
        def str(self, flows = None):
            if flows:
                return ', '.join(str(flows[item]) for item in self.keys())
            else:
                return ', '.join(str(item) if rate is None else f'{item} @ {rate}' for item,rate in self.items())
        __str__ = str
        def __setitem__(self, item, rate):
            raise NotImplementedError
        def _jsonObj(self):    
            from .jsonconv import _jsonObj
            return {k.name: _jsonObj(v) for k,v in self.items()}

    class Outputs(_ExternalFlows):
        __slots__ = ()
        def __setitem__(self, item, rate):
            item = asItem(item)
            if rate is None:
                return dict.__setitem__(self, item, None)
            rate = frac(rate)
            if rate < 0:
                raise ValueError
            return dict.__setitem__(self, item, rate)


    class Inputs(_ExternalFlows):
        __slots__ = ()
        def __setitem__(self, item, rate):
            item = asItem(item)
            if rate is None:
                return dict.__setitem__(self, item, None)
            rate = frac(rate)
            if rate > 0:
                rate = -rate
            return dict.__setitem__(self, item, rate)
    

    class _Dict(dict):
        __slots__ = ()
        def __init__(self, vals = None):
            if vals is None:
                return
            for k, v in vals.items() if isinstance(vals, Mapping) else vals:
                self[k] = v

    class Constraints(_Dict):
        __slots__ = ()
        def __str__(self):
            def fmtTerm(num, item):
                if num == 1:
                    return f'{item}'
                elif num == -1:
                    return f'-{item}'
                else:
                    return f'{item}*{rate}'
            return ', '.join(str(item) if rate is None
                             else f'{item} >= {rate}' if isinstance(rate, Number)
                             else f'{item} = {fmtTerm(*rate)}'
                             for item,rate in self.items())
        def __setitem__(self, item, rate):
            item = asItem(item)
            if isinstance(rate, Ingredient):
                rate = (1, rate)
            if type(rate) is tuple:
                if len(rate) != 2:
                    raise AttributeError('rate tuple length must be 2')
                rate = (frac(rate[0]), asItem(rate[1]))
            else:
                rate = frac(rate)
            return dict.__setitem__(self, item, rate)
        def _jsonObj(self):
            from .jsonconv import _jsonObj
            return {k.name: _jsonObj(v) for k,v in self.items()}

    class Priorities(_Dict):
        __slots__ = ()
        def __str__(self):
            def tostr(k):
                return str(k) if isinstance(k, Ingredient) else f'recipe {k.name}'
            return ', '.join(f'ignore {tostr(k)}' if p <= IGNORE else f'{tostr(k)}: {p}' for k, p in self.items())
        def __setitem__(self, key, priority):
            priority = frac(priority)
            if priority < Box.MIN_PRIORITY or priority > Box.MAX_PRIORITY:
                raise ValueError('priority must be between {Box.MIN_PRIORITY} and {Box.MAX_PRIORITY}, inclusive')
            return dict.__setitem__(self, key, priority)
        def _jsonObj(self):
            from .jsonconv import _jsonObj
            return {f"{'i' if isinstance(k, Ingredient) else 'r'} {k.name}": _jsonObj(v) for k,v in self.items()}

    name: str
    inner: Group
    outputs: Outputs
    inputs: Inputs
    constraints: Constraints
    priorities: Priorities

    def __init__(self, inner, *, name = None,
                 outputs = None, extraOutputs = (), outputTouchups = {}, outputsLoose = False,
                 inputs = None, extraInputs = (), inputTouchups = {}, inputsLoose = True,
                 constraints = None, priorities = None,
                 allowExtraInput = False):
        """Create a new box.

        A box is a wrapper around a group with additional constraints to
        limit flows.  For example it is rare for satellite machine to be
        running 100% of the time.  This means that simply estimating the
        maximum flows will not give you an accurate picture of the resources
        needed, by grouping the satellite with a rocket silo, wrapping it
        in a box, and then solving it, you can can a better idea of the
        resources needed for creating space science packs.  After solving,
        the satellite machine will be throttled to produce just enough
        satellites for the rocket silo.

        Boxes can nest inside one another.

        Also see `box()`, which is a shortcut to both create a box and then
        solve it.

        *inner*
            A `Group`.

        *name*
            A name for the box

        *outputs*, *extraOutputs*, *outputTouchups*, *inputs*, *extraInputs*, *inputTouchups*
            A sequence or mapping of outputs and inputs.  The key value can
            either be a rate in seconds or None.  If a list, elements can
            either be an item, a tuple of the form (item, rate) (as generated
            with the `@` operator), or a flow.

            If *outputs* or *inputs* is unspecified than they will be derived
            automatically based on if the flow is internal or not.  The
            *outputTouchups* and *inputTouchups* can be used to adjust the inputs
            or outputs afterwards.

            *extraOutputs* or *extraInputs* can be used to specify auxiliary
            flows, which have special meaning to the solver.

        *outputsLoose* 
            If True (default False) than all output rates will
            become constraints

        *inputsLoose* 
            If True (the default) than all input rates will become
            constraints
        
        *constraints*
            A mapping of extra constraints for the solver.  The key
            is an item.  If the value is a rate, with input rates being
            negative, than that flow will need to be at least the given rate.
            If the value is another item or a tuple of the form (num, item)
            that then a constraints is added of the form::

              <key-item> = <num> * <item>

        *priority* 
            A mapping of priories for the solver.  The key is either a
            recipe or an item.  The value in a number between -100 and 100
            with larger values having a higher priority.  The default priority
            is 0.  The constant `IGNORE` can be used for the lowest priority
            (-100) and has special meaning to the solver.

        *allowExtraInputs*
            |nbsp|

        """
        from .solver import SolveRes
        if not isinstance(inner, Group):
            inner = Group(inner)
        self.inner = inner
        self.name = name
        self.outputs = Box.Outputs(outputs)
        self.inputs = Box.Inputs(inputs)
        self.priorities = Box.Priorities(priorities)
        self.constraints = Box.Constraints(constraints)
        if outputs is None or inputs is None:
            inputs_ = set()
            outputs_ = set()
            for m in self.inner.flatten():
                inputs_ |= m.inputs.keys()
                outputs_ |= m.outputs.keys()
            common = inputs_ & outputs_
            if outputs is None:
                self.outputs = Box.Outputs(outputs_ - common)
            if inputs is None:
                if allowExtraInput:
                    self.inputs = Box.Inputs(inputs_ - self.outputs.keys())
                    for item in self.inputs.keys() & common:
                        self.priorities[item] = IGNORE
                else:
                    self.inputs = Box.Inputs(inputs_ - common)
        mainOutput = [item for item in self.outputs if item is not itm.empty_barrel]
        if len(mainOutput) == 1 and self.name is None:
            self.name = 'b-{}'.format(mainOutput[0].name)
        for item in extraOutputs:
            self.outputs[item] = None
            self.priorities[item] = IGNORE
        for item in extraInputs:
            self.inputs[item] = None
            self.priorities[item] = IGNORE
        for item, rate in Box.Outputs(outputTouchups).items():
            self.outputs[item] = rate
        for item, rate in Box.Inputs(inputTouchups).items():
            self.inputs[item] = rate
        if outputsLoose:
            for item, rate in self.outputs.items():
                if rate is not None:
                    self.constraints[item] = rate
                    self.outputs[item] = None
        if inputsLoose:
            for item, rate in self.inputs.items():
                if rate is not None:
                    self.constraints[item] = rate
                    self.inputs[item] = None

    def _jsonObj(self, objs, **kwargs):
        from .jsonconv import _jsonObj
        if id(self) in objs:
            return objs[id(self)]
        obj = {}
        obj['name'] = self._fallbackName
        obj['id'] = objs.add(id(self))
        if self.name:
            obj['label'] = self.name
        obj['inner'] = self.inner._jsonObj(objs = objs, **kwargs)
        for k in ['outputs', 'inputs', 'constraints', 'priorities']:
            v = getattr(self,k)
            if v:
                obj[k] = v._jsonObj()
        return obj

    def summarize(self):
        """:meta private:"""
        obj = _copy(self)
        obj.inner = obj.inner.summarize()
        return obj
    
    def summary(self, out = None, *, prefix = '', includeSolvedBoxFlows = True, includeMachineFlows = False, includeBoxDetails = True, flowsItemFilter = None):
        self._summary(out, prefix, includeSolvedBoxFlows, includeMachineFlows, includeBoxDetails, flowsItemFilter)

    def _summary(self, out, prefix, includeSolvedBoxFlows, includeMachineFlows, includeBoxDetails, flowsItemFilter, _namePrefix = ''):
        if out is None:
            out = sys.stdout
        out.write(prefix)
        nameSuffix = ''
        if getattr(self, 'throttle', 1) != 1:
            nameSuffix = f' @{self.throttle:.6g}'
        if self.name is None:
            out.write(f'{_namePrefix}Box{nameSuffix}:\n')
        else:
            out.write(f'{_namePrefix}{self.name}{nameSuffix}:\n')
        prefix = prefix + '  '
        self.inner._summary(out, prefix + '  ', includeSolvedBoxFlows, includeMachineFlows, includeBoxDetails, flowsItemFilter)
        flows = self.flows()
        if flows.state != FlowsState.OK:
            out.write(f'{prefix}{flows.state.name}\n')
        if flows.state.ok() and includeSolvedBoxFlows:
            pass
        else:
            flows = None
        out.write(f'{prefix}Outputs: {self.outputs.str(flows)}\n')
        out.write(f'{prefix}Inputs: {self.inputs.str(flows)}\n')
        if self.constraints:
            out.write(f'{prefix}Constraints: {self.constraints}\n')
        if self.priorities:
            out.write(f'{prefix}Priorities: {self.priorities}\n')

    def solved(self):
        return self.flows().state == FlowsState.OK

    def solve(self):
        """Solve the box so all contraints are met.

        Returns a `SolveRes` enum with the result.  See the `SolveRes`
        documenation for meaning of the enum values.

        """
        from .solver import SolveRes
        solver = self.solver()
        solver.solve()
        res, _ = solver.apply()
        return res

    def solver(self, **kwargs):
        """Return a solver instance for debugging or more control over the solve process."""
        from . import solver
        return solver.Solver(solver.LinearEqSystem.fromBox(self, **kwargs))

    def _flows(self, throttle, _includeInner):
        res = _MutableFlows()
        orig = self.inner._flows(throttle = None, _includeInner = False)
        state = orig.state
        if throttle is None:
            throttle = 1
        for item,rate in self.outputs.items():
            flow = orig[item]
            underflow = flow.underflow
            annotation = ''
            if flow.rate() < 0:
                state = max(state, FlowsState.UNSOLVED)
                annotation = '!'
            elif rate is not None and flow.rate() < rate:
                underflow = True
            elif item in self.constraints and isinstance(self.constraints[item], Number) and flow.rate() < self.constraints[item]:
                underflow = True
            elif rate is not None and flow.rate() > rate:
                state = max(state, FlowsState.UNSOLVED)
                annotation = '*'
            if underflow:
                annotation = '!'
                state = max(state, FlowsState.UNDERFLOW)
            if _includeInner:
                res.byItem[item] = flow.copy(factor = throttle, adjusted = False, underflow = underflow, annotation = annotation)
            else:
                res.byItem[item] = Flow(item, rateOut = flow.rate() * throttle, underflow = underflow, annotation = annotation)
        for flow in orig:
            if flow.item in self.outputs or flow.item in self.inputs or flow.item is itm.electricity:
                continue
            if flow.rateIn == 0 == flow.rateOut:
                continue
            annotation = ''
            if flow.rate() < 0:
                state = max(state, FlowsState.UNSOLVED)
                annotation = '!'
            elif flow.rate() > 0:
                state = max(state, FlowsState.UNSOLVED)
                annotation = '*'
            if _includeInner:
                res.byItem[flow.item] = flow.copy(factor = throttle, adjusted = False, annotation = annotation)
            else:
                pass
        for item,rate in self.inputs.items():
            flow = orig[item]
            annotation = ''
            if -flow.rate() < 0 or (rate is not None and -flow.rate() < rate):
                state = max(state, FlowsState.UNSOLVED)
                annotation = '!'
            elif item in self.constraints and flow.rate() < self.constraints[item]:
                state = max(state, FlowsState.UNSOLVED)
                annotation = '!'
            elif rate is not None and -flow.rate() > rate:
                state = max(state, FlowsState.UNSOLVED)
                annotation = '*'
            if flow.rateIn == 0 == flow.rateOut and annotation == '':
                continue
            if _includeInner:
                res.byItem[item] = flow.copy(factor = throttle, adjusted = False, annotation = annotation)
            else:
                res.byItem[item] = Flow(item, rateIn = -flow.rate() * throttle, underflow = underflow, annotation = annotation)                
        res.byItem[itm.electricity] = orig[itm.electricity]
        if not _includeInner and not state.ok():
            res.byItem.clear()
        res.state = state
        return BoxFlows(res)

    def find(self, **kwargs):
        return self.inner.find(**kwargs)
    
    def resetThrottle(self):
        self.inner.resetThrottle()

    def finalize(self, *, roundUp = True, recursive = True, _res = None):
        """Finalize the result.

        Finalize the result by converting `UnboundedBox`'es into a normal ones
        and if *roundUp* is True (the default) also round up multiples of
        machines and adjusting the throttle to compensate.  If *recursive* is
        True (default False) than recurse into any inner boxes, otherwise
        the contents of inner boxes are left alone.  `UnboundedBox`'es one
        level deep are always converted.

        Returns an instance of `FinalizeResult`.

        """
        res = FinalizeResult() if _res is None else _res
        res.factory = self
        _finalizeInner(self.inner, roundUp, recursive)
        flows = self.flows()
        filteredOutputs = Box.Outputs()
        for item, rate in self.outputs.items():
            if self.priorities.get(item,None) == IGNORE:
                if flows[item].rate() == 0:
                    del self.priorities[item]
                    continue
                res.extraOutputs.append(item)
            filteredOutputs[item] = rate
        self.outputs = filteredOutputs
        filteredInputs = Box.Inputs()
        for item, rate in self.inputs.items():
            if flows[item].rate() == 0:
                res.unusedInputs.append(item)
            else:
                filteredInputs[item] = rate
        self.inputs = filteredInputs
        if self.__class__ is UnboundedBox:
            self.__class__ = Box
        return res

    def finalizeAll(self, roundUp = True):
        """Finalize everything.

        Equivalent to ``self.finalize(roundUp, recursive = True)``
        """
        return self.finalize(roundUp = roundUp, recursive = True)

class UnboundedBox(Box):
    """A special type of box for use with the solver.
        
    When used with the solver, the solver will adjust the number of machines
    rather then the machine's throttle.

    The inner parameter must be a machine or a group of the form::

       Group(Mul(1,machine1), Mul(1,machine2), ...)

    i.e::

       1*machine1 + 1*machine2 + ...

    The inner `Mul` is important to give the solver a number to adjust.  The
    actual number is unimportant.

    An UnboundedBox can nest instead a normal box, which is useful to let
    the solver tell you the number of machines you need to support another
    machine.
    """
    _fallbackName = '<unbounded box>'
    def __init__(self, inner, **kwargs):
        """Create a new `UnboundedBox`, the parameters are the same as `Box`."""
        if not isinstance(inner, (Group, Mul)):
            inner = Group(Mul(1,inner))
        super().__init__(inner, **kwargs)
    def scale(self, factor):
        """Adjust the number of machines uniformly."""
        factor = frac(factor)
        for m in self.inner:
            m.num *= factor

class BlackBox(BoxBase):
    """Wrap a box to hide it's internals.

    When used with the solver, the solver will treat the black box as a
    simple machines and throttle all internal machines uniformity.  This
    allows them to be used inside an unbounded box and can also simplify the
    solve process.

    """
    _fallbackName = '<black box>'
    def __init__(self, box, *, name = None):
        self.throttle = 1
        self._inner = box
        self.name = name if name is not None else self._inner.name
 
    def _flows(self, throttle, _includeInner):
        if throttle is None:
            throttle = self.throttle
        return self._inner._flows(throttle, False)

    @property
    def inputs(self):
        return self._inner.inputs

    @property
    def outputs(self):
        return self._inner.outputs

    def resetThrottle(self):
        self.throttle = 1

def _finalizeInner(m, roundUp, recursive):
    if isinstance(m, Box) and recursive:
        m.finalize(roundUp = roundUp, recursive = True)
    elif isinstance(m, Group):
        _finalizeGroup(m, roundUp, recursive)
    elif isinstance(m, Mul) and isinstance(m.machine, (Machine, BlackBox)):
        throttle = m.num * m.machine.throttle
        if throttle == 0: return None
        if roundUp:
            m.num = ceil(throttle)
            m.machine.throttle = div(throttle, m.num)
        else:
            m.num = throttle
            m.machine.throttle = 1
    elif isinstance(m, Mul):
        m0 = _finalizeInner(m.machine, roundUp, recursive)
        if m0 is None: return None
        m.machine = m0
    return m

def _finalizeGroup(grp, roundUp, recursive):
    machines = []
    for m in grp:
        wasUnboundedBox = isinstance(m, UnboundedBox)
        m = _finalizeInner(m, roundUp, recursive)
        if m is None: continue
        # fixme: if recursive and wasUnboundedBox with a single machine
        # and no constraints then remove the box as is now unnecessary
        machines.append(m)
    grp.machines = machines

@dataclass
class FinalizeResult:
    factory: Box = None
    extraOutputs: list = field(default_factory=list)
    unusedInputs: list = field(default_factory=list)

class CantFinalizeError(ValueError):
    pass
    



from __future__ import annotations
from dataclasses import dataclass,field
from collections import defaultdict
from typing import NamedTuple
from itertools import chain
# from numbers import Rational
from collections.abc import Sequence,Iterator,Mapping
from enum import Enum
from .ordenum import OrdEnum
from .fracs import Frac,frac,Inf,div,diva,isfinite,isnan,isinf
from copy import copy
import sys
from . import itm, rcp, rcpinst

__all__ = ('Default',
           'Ingredient', 'Item', 'Research', 'Electricity', 'Module', 'FakeModule',
           'MachineBase', 'Machine', 'CraftingMachine', 'Mul', 'Group',
           'Flow', 'OneWayFlow', 'FlowsState', 'Flows', 'EffectBase', 'Effect', 'Bonus', 'Mode',
           'Rcp', 'Recipe', 'RecipeComponent', 'IGNORE', 'InvalidModulesError')

class Uniq:
    def __copy__(self):
        return self
    def __deepcopy__(self,memo):
        return self
    def __eq__(self, other):
        # we want to compare by identity but also give other.__eq__ a chance
        # to be called if defined, so return NotImplemented...
        return NotImplemented 
    def __ne__(self, other):
        return NotImplemented
    def __hash__(self):
        return object.__hash__(self)

class Immutable:
    def __setattr__(self, name, value):
        raise TypeError
    def __delattr__(self, name):
        raise TypeError

class DefaultType(Uniq):
    __slots__ = ()
    def __repr__(self):
        return 'Default'
    def __new__(cls):
        return DEFAULT
    def __copy__(self):
        return self
    def __deepcopy__(self, _):
        return self

Default = object.__new__(DefaultType)
    
class Ingredient(Uniq,Immutable):
    """Base class for all items."""
    __slots__ = ('name', 'order')
    def __init__(self, name, order):
        object.__setattr__(self, 'name', name)
        object.__setattr__(self, 'order', order)
    def __str__(self):
        return self.name
    def __repr__(self):
        from ._helper import toPythonName
        pythonName = toPythonName(self.name)
        if getattr(itm, pythonName, None) is self:
            return 'itm.'+toPythonName(self.name)
        else:
            return f'<{type(self).__name__}: {self.name}>'
    def __matmul__(self, rate):
        return (self, rate)
        
    # def __lt__(self, other):
    #     if not isinstance(other, Ingredient):
    #         return NotImplemented
    #     return self.order < other.order
    # def __le__(self, other):
    #     return self.__eq__(other) or self.__lt__(other)
    # def __gt__(self, other):
    #     if not isinstance(other, Ingredient):
    #         return NotImplemented
    #     return self.order > other.order
    # def __ge__(self, other):
    #     return self.__eq__(other) or self.__gt__(other)

class Item(Ingredient):
    __slots__ = ('stackSize', 'fuelValue', 'fuelCategory')
    def __init__(self, name, order, stackSize, fuelValue = 0, fuelCategory = ''):
        super().__init__(name, order)
        object.__setattr__(self, 'stackSize', stackSize)
        object.__setattr__(self, 'fuelValue', fuelValue)
        object.__setattr__(self, 'fuelCategory', fuelCategory)

class Research(Ingredient):
    __slots__ = ()
    pass

class Electricity(Ingredient):
    __slots__ = ()
    pass

class Module(Item):
    __slots__ = ('effect','limitation')
    def __init__(self,name,order,stackSize,effect,limitation=None,limitationBlacklist=None):
        super().__init__(name,order,stackSize)
        object.__setattr__(self, 'effect', effect)
        object.__setattr__(self, 'limitation', limitation)
    def _jsonObj(self):
        return self.name
    def __mul__(self, num):
        return num*[self]
    __rmul__ = __mul__

class FakeModule(Module):
    def __init__(self, speed=0, productivity=0, consumption=0, pollution=0):
        super().__init__('fake-module', '', 0, Bonus(speed, productivity, consumption, pollution))
    def __repr__(self):
        return 'FakeModule({})'.format(', '.join(f'{k}={getattr(self.effect,k)!r}' for k in self.effect._fields if getattr(self.effect,k) != 0))
    def __str__(self):
        return f'module: {self.effect}'
    def _jsonObj(self):
        from .jsonconv import _jsonObj
        return {k: _jsonObj(getattr(self.effect, k)) for k in self.effect._fields if getattr(self.effect,k) != 0}
    def __eq__(self, other):
        return type(self) == type(other) and self.effect == other.effect
    def __ne__(self, other):
        return type(self) != type(other) or  self.effect != other.effect
    def __hash__(self, other):
        raise NotImplementedError

@dataclass
class MachineBase:
    def _flows(self, throttle, _includeInner):
        raise NotImplementedError

    def flows(self, throttle = None):
        return self._flows(throttle, True)

    def flow(self, item, throttle = None):
        return self.flows(throttle).flow(item)

    def __mul__(self, fac):
        return Mul(self,fac)

    def __rmul__(self, fac):
        return Mul(self,fac)

    def __add__(self, other):
        return Group(self,other)

    def __new__(cls, *args, **kwargs):
        if getattr(cls, 'abstract', None) == cls:
            raise TypeError(f"can't create {cls.__name__}")
        return object.__new__(cls)

@dataclass(init=False)
class Machine(MachineBase):
    """A entity that used directly or indirectly to produce something."""
    throttle: Rational
    blueprintInfo: dict = field(default = None, init = False, repr = False, compare = False)
    __flows1: Flows = field(default = None, init = False, repr = False, compare = False)
    __flows: Flows = field(default = None, init = False, repr = False, compare = False)

    @property
    def machine(self):
        return self

    @property 
    def num(self): 
        return 1

    @property
    def recipe(self):
        return None

    def resetThrottle(self):
        self.throttle = 1
    
    def __setattr__(self, prop, val):
        if prop == 'throttle':
            val = frac(val)
            if val < 0 or val > 1:
                raise ValueError('throttle must be between 0 and 1, inclusive')
        if prop != '_Machine__flows' and  prop != '_Machine__flows1' and prop != 'blueprintInfo':
            self.__flows = None
            if prop != 'throttle':
                self.__flows1 = None
        super().__setattr__(prop, val)

    def _flatten(self, lst, factor):
        if factor == 1:
            lst.append(self)
        else:
            lst.append(Mul(self, factor))

    def summarize(self):
        return self

    def __str__(self):
        name = type(self).__name__
        parts = []
        if self.recipe:
            parts.append(self.recipe.name)
        if self.throttle != 1:
            parts.append(f'@{self.throttle:g}')
        modulesStr = self._modulesStr()
        if modulesStr:
            parts.append(modulesStr)
        if parts:
            return '{}({})'.format(name, '; '.join(parts))
        else:
            return name

    def _modulesStr(self):
        return ''

    def _jsonObj(self, objs, **kwargs):
        from .jsonconv import _jsonObj
        if id(self) in objs:
            return objs[id(self)]
        obj = {}
        obj['name'] = self.name
        obj['id'] = objs.add(id(self))
        if self.throttle and self.throttle != 1:
            obj['throttle'] = _jsonObj(self.throttle)
        if self.recipe is not None:
            obj['recipe'] = self.recipe._jsonObj(**kwargs)
        return obj

    def flatten(self):
        return self

    @property
    def inputs(self):
        return {flow.item: None for flow in self.flows(1) if flow.rate() < 0 and flow.item is not itm.electricity}

    @property
    def outputs(self):
        return {flow.item: None for flow in self.flows(1) if flow.rate() > 0}

    def _calc_flows(self, throttle):
        return _MutableFlows()

    def _flows(self, throttle, _includeInner):
        if throttle is None:
            throttle = self.throttle
        if throttle == 1 and self.__flows1 is not None:
            return self.__flows1
        if throttle == self.throttle and self.__flows is not None:
            return self.__flows
        res = self._calc_flows(throttle)
        if isinstance(res, _MutableFlows):
            res = SimpleFlows(res)
        if throttle == 1:
            self.__flows1 = res
        if throttle == self.throttle:
            self.__flows = res
        return res

    def __init__(self, *, throttle = 1):
        self.throttle = throttle

    def bonus(self) -> Bonus:
        return Bonus()

    def energyUsage(self, throttle):
        if throttle is None:
            throttle = self.throttle
        return self.energyDrain + throttle * self.baseEnergyUsage * (1 + self.bonus().consumption)

@dataclass(init = False)
class CraftingMachine(Machine):
    """An entity that produce something."""
    recipe: Recipe = None
    name = 'crafting-machine'
    craftingSpeed = 1
    baseEnergyUsage = 0
    energyDrain = 0
    pollution = 0

    def __init__(self, recipe = None, **kws):
        super().__init__(**kws)
        self.recipe = recipe

    def __setattr__(self, prop, val):
        if prop == 'recipe':
            if val is None or isinstance(val, Recipe):
                pass
            elif isinstance(val, Rcp):
                val = val.inst()
            elif isinstance(val, str):
                val = rcp.byName[str].inst()
            else:
                raise TypeError(f'unexpected type for recipe: {type(val)}')
        super().__setattr__(prop, val)

    def _calc_flows(self, throttle):
        flows = _MutableFlows()
        if self.recipe is None: return flows
        numInOut = defaultdict(lambda: [0,0])
        for rc in self.recipe.inputs:
            numInOut[rc.item][0] = rc.num
        for rc in self.recipe.outputs:
            numInOut[rc.item][1] = rc.num
        b = self.bonus()
        time = diva(self.recipe.time, self.craftingSpeed, 1 + b.speed)
        for item, (numIn, numOut) in numInOut.items():
            bonusOut = 0
            if numOut > numIn:
                bonusOut = (numOut-numIn)*b.productivity
            rateIn = div(numIn, time)
            rateOut = div(numOut + bonusOut, time)
            flows.addFlow(item, rateIn = throttle*rateIn, rateOut = throttle*rateOut, adjusted = throttle != 1)
        return flows

@dataclass(repr=False)
class Mul(MachineBase):
    """A group of identical machines.

    num must be positive but fractional values are permitted.

    """
    num: Rational
    machine: MachineBase

    def __init__(self, *args):
        if isinstance(args[0], MachineBase):
            self.machine = args[0]
            self.num = args[1]
        else:
            self.num = args[0]
            self.machine = args[1]

    def __setattr__(self, prop, val):
        if prop == 'num':
            val = frac(val)
            if val < 0:
                raise ValueError('Mul.num must be positive')
        super().__setattr__(prop, val)

    def __str__(self):
        return f'{self.num} {self.machine}'

    def __repr__(self):
        return f'Mul({self.num!r}, {self.machine!r})'

    def _jsonObj(self, objs, **kwargs):
        from .jsonconv import _jsonObj
        if id(self) in objs:
            return objs[id(self)]
        obj = {}
        obj['name'] = '<mul>'
        obj['id'] = objs.add(id(self))
        obj['num'] = _jsonObj(self.num)
        obj['machine'] = self.machine._jsonObj(objs = objs, **kwargs)
        return obj

    def __repr__(self):
        return repr(self.num) + '*' + repr(self.machine)

    @property
    def inputs(self):
        return self.machine.inputs

    @property
    def outputs(self):
        return self.machine.outputs

    @property
    def recipe(self):
        return self.machine.recipe

    def resetThrottle(self):
        self.machine.resetThrottle()

    def _flows(self, throttle, _includeInner):
        return self.machine._flows(throttle, _includeInner) * self.num

    def _flatten(self, lst, num):
        self.machine._flatten(lst, num * self.num)

    def flatten(self):
        lst = []
        self._flatten(lst, 1)
        if isinstance(self.machine,Group) or len(lst) > 1:
            return Group(lst)
        else:
            return lst[0]

    def summarize(self):
        return Mul(self.num, self.machine.summarize())
        

@dataclass(repr=False,init=False)
class Group(Sequence,MachineBase):
    """A group of machines."""
    machines: List[MachineBase]
    def __init__(self, *machines):
        super().__init__()
        if len(machines) == 1 and isinstance(machines[0], Sequence) and not isinstance(machines[0], Group):
            self.machines = list(machines[0])
        elif len(machines) == 1 and isinstance(machines[0], Iterator):
            self.machines = list(machines[0])
        else:
            self.machines = list(machines)

    @property
    def machine(self):
        return self

    @property 
    def num(self): 
        return 1
    
    def _flatten(self, lst, num):
        for m in self.machines:
            m._flatten(lst, num)

    def flatten(self):
        lst = []
        self._flatten(lst, 1)
        return Group(lst)

    def __iter__(self):
        return iter(self.machines)

    def __getitem__(self, idx):
        return self.machines[idx]

    def __len__(self):
        return len(self.machines)

    def __repr__(self):
        return 'Group(' + ', '.join(repr(m) for m in self) + ')'

    def __str__(self):
        res = '('
        res += ' + '.join(str(m) for m in self.machines)
        res += ')'
        return res

    def _jsonObj(self, objs, **kwargs):
        from .jsonconv import _jsonObj
        if id(self) in objs:
            return objs[id(self)]
        obj = {}
        obj['name'] = '<group>'
        obj['id'] = objs.add(id(self))
        obj['machines'] = [m._jsonObj(objs = objs, **kwargs) for m in self.machines]
        return obj

    def __add__(self, other):
        return Group(*self,other)

    def summary(self, out = sys.stdout, *, prefix = '',
                includeSolvedBoxFlows = True, includeMachineFlows = True, includeBoxDetails = True,
                flowsItemFilter = None):
        self._summary(out, prefix, includeSolvedBoxFlows, includeMachineFlows, includeBoxDetails, flowsItemFilter)

    def summarize(self):
        from .box import BoxBase
        class Tally:
            __slots__ = ('num', 'rateIn', 'rateOut', 'consumption', 'pollution', 'throttle')
            def __init__(self):
                self.num = 0
                self.rateIn = 0
                self.rateOut = 0
                self.consumption = 0
                self.pollution = 0
                self.throttle = 0
        tally = defaultdict(Tally)
        for m in self:
            num = 1
            if isinstance(m, Mul):
                num = m.num
                m = m.machine
            if isinstance(m, BoxBase):
                if id(m) in tally:
                    tally[id(m)].num += num
                else:
                    tally[id(m)] = Mul(num, m.summarize())
            elif isinstance(m, Group):
                if id(m) in tally:
                    tally[id(m)].num += num
                else:
                    tally[id(m)] = Mul(num, m.summarize())
            else:
                recipe = m.recipe
                fuel = m.fuel if hasattr(m, 'fuel') else None
                x = tally[(type(m), recipe, fuel)]
                x.num += num
                x.throttle += num*m.throttle
                if recipe is None: continue
                # We can't just average the speed and productivity bonus, so,
                # instead we average the rateIn and rateOut for abstract items
                # in which the machine's recipe produces and consumes exactly
                # one of.  We them derive the required speed and productivity
                # bonus from those numbers.
                b = m.bonus()
                t = diva(m.recipe.time, m.craftingSpeed, 1 + b.speed, 1 + b.productivity) + getattr(m, 'delay', 0)
                x.rateIn += diva(1, t, 1 + b.productivity)
                x.rateOut += div(1, t)
                x.consumption += b.consumption
                x.pollution += b.pollution
        grp = []
        for k, v in tally.items():
            if type(k) is int:
                if v.num == 1:
                    grp.append(v.machine)
                else:
                    grp.append(v)
            else:
                cls, recipe, fuel = k
                num = v.num
                m = cls()
                if v.throttle != 1:
                    m.throttle = div(v.throttle, num)
                if recipe:
                    m.recipe = recipe
                if fuel:
                    m.fuel = fuel
                # Given:
                #   time = recipeTime / craftingSpeed / (1 + speed) / (1 + productivity) + delay
                #   rateIn = 1 / time / (1 + productivity)
                #   rateOut = 1 / time
                # To derive the required speed and productivity we 
                # need to solve for speed and productivity
                if v.rateIn > 0 and v.rateOut > 0:
                    craftingSpeed, delay = m.craftingSpeed, getattr(m, 'delay', 0)
                    rateIn, rateOut = div(v.rateIn, num), div(v.rateOut, num)
                    speed = div(craftingSpeed - craftingSpeed*delay*rateOut - rateIn*recipe.time, craftingSpeed*(delay*rateOut - 1))
                    productivity = div(rateOut - rateIn, rateIn)
                    consumption = div(v.consumption, num)
                    pollution = div(v.pollution, num)
                    if speed != 0 or productivity != 0 or consumption != 0 or pollution != 0:
                        m.modules = [FakeModule(speed, productivity, consumption, pollution)]
                if num > 1:
                    m = Mul(m, num)
                grp.append(m)
        return Group(grp)
 
    def _summary(self, out, prefix, includeSolvedBoxFlows, includeMachineFlows, includeBoxDetails, flowsItemFilter):
        from .box import BoxBase, Box
        byRecipe = defaultdict(list)
        grpNum = 0
        for m in self.machines:
            if isinstance(m.machine, Group) or isinstance(m.machine, BoxBase):
                byRecipe[grpNum] = m
                grpNum += 1
            elif m.machine.recipe:
                byRecipe[m.machine.recipe].append(m)
            else:
                byRecipe[type(m.machine)].append(m)
        prevIsGroup = False
        for key, val in byRecipe.items():
            orig = val
            namePrefix = ''
            if isinstance(val, Mul):
                namePrefix = f'{val.num:.3g}x '
                val = val.machine
            if isinstance(val, Group):
                out.write(prefix)
                out.write(f'{namePrefix}Group: \n')
                val._summary(out, prefix + '    ', includeSolvedBoxFlows, includeMachineFlows, includeBoxDetails, flowsItemFilter)
                prevIsGroup = True
            elif isinstance(val, BoxBase):
                if includeBoxDetails and isinstance(val, Box):
                    val._summary(out, prefix, includeSolvedBoxFlows, includeMachineFlows, includeBoxDetails, flowsItemFilter, namePrefix)
                else:
                    out.write(f'{prefix}{namePrefix}{val}')
                    if includeSolvedBoxFlows or includeMachineFlows:
                        flows = orig.flows()
                        if flows.state.ok():
                            out.write(':\n')
                            out.write('{}         {}\n'.format(prefix, flows.filter(items = flowsItemFilter)))
                        else:
                            out.write('\n')
                    else:
                        out.write('\n')
            elif isinstance(key, Recipe):
                if (prevIsGroup and includeMachineFlows):
                    out.write('{}\n'.format(prefix))
                num = 0
                byMachine = {}
                for m in val:
                    num += m.num
                    x = byMachine.setdefault(type(m.machine), {'num': 0, 'bonus': Bonus(), 'throttle': 0})
                    x['num'] += m.num
                    x['bonus'] += m.num*m.machine.bonus()
                    x['throttle'] += m.num*m.machine.throttle
                g = Group(val)
                out.write('{}{: >4.3g}x {}: '.format(prefix, num, key.name))
                pos = 0
                for k, v in byMachine.items():
                    if v['num'] ==  0:
                        continue
                    if pos > 0:
                        out.write('; ')
                    if v['num'] != num:
                        out.write('{}x '.format(v['num']))
                    out.write(k.name)
                    throttle = div(v['throttle'], v['num'])
                    if (throttle != 1):
                        out.write('  @{:.6g}'.format(throttle))
                    if x['bonus']:
                        out.write('  {}'.format(x['bonus'] / v['num']))
                    pos += 1
                if includeMachineFlows:
                    out.write(':\n')
                    out.write('{}         {}\n'.format(prefix, g.flows().filter(items = flowsItemFilter)))
                else:
                    out.write('\n')
                prevIsGroup = False
            else:
                num = 0
                for m in val:
                    num += m.num
                out.write('{}{: >4.3g}x {}\n'.format(prefix, num, key.__name__))
                prevIsGroup = False

    def _flows(self, throttle, _includeInner):
        res = _MutableFlows()
        state = FlowsState.OK
        for m in self.flatten():
            flows = m._flows(throttle, _includeInner)
            for flow in flows:
                res.merge(flow)
            if not flows.state.ok():
                state = FlowsState.INVALID
        res.reorder()
        res.state = state
        return NetFlows(res)

    def setThrottle(self, throttle, filter = lambda _: True):
        for m in self.flatten():
            if filter(m.machine):
                m.machine.throttle = throttle

    def adjThrottle(self, factor, filter = lambda _: True):
        seen = set()
        for m in self.flatten():
            m = m.machine
            if filter(m) and id(m) not in seen:
                m.machine.throttle *= factor
                seen.add(id(m))

    def resetThrottle(self):
        for m in self.flatten():
            m.resetThrottle()

    def find(self, *,
             input = None, output = None, recipe = None, machine = None,
             inputs = (), outputs = (), recipes = (), machines = ()):
        if input is not None:
            if inputs:
                raise ValueError('both input and inputs can not be defined')
            inputs = {input}
        else:
            inputs = set(inputs)
        if output is not None:
            if outputs:
                raise ValueError('both output and outputs can not be defined')
            outputs = {output}
        else:
            outputs = set(outputs)
        if recipe is not None:
            if recipes:
                raise ValueError('both recipe and recipes can not be defined')
            recipes = {recipe}
        else:
            recipes = set(recipes)
        if machine is not None:
            if machines:
                raise ValueError('both machine and machines can not be defined')
            machines = {machine}
        else:
            machines = tuple(machines)
        res = []
        for m in self.machines:
            if not inputs.isdisjoint(m.inputs):
                res.append(m)
            elif not outputs.isdisjoint(m.outputs):
                res.append(m)
            elif m.recipe in recipes:
                res.append(m)
            elif machines and isinstance(m.machine, machines):
                res.append(m)
        return Group(res)

class Mode(Enum):
    NORMAL = 1
    EXPENSIVE = 2

def _unitForItem(item):
    from . import config
    for cls, unit in config.displayUnit.get():
        if cls is None or isinstance(item, cls):
            return unit
    raise ValueError

def _fmt_rate(item, rate, unit):
    if unit is None:
        from .units import UNIT_SECONDS
        unit = UNIT_SECONDS
    if type(rate) is int or isfinite(rate):
        if callable(unit.conv):
            rate = unit.conv(item, rate)
        else:
            rate = rate * unit.conv
        return '{:,.6g}{}{}'.format(rate,unit.sep,unit.abbr)
    elif isnan(rate):
        return 'Unknown'
    elif isinf(rate):
        if rate < 0:
            return 'Negative'
        else:
            return 'Positive'
    else:
        raise ValueError

class OneWayFlow(NamedTuple):
    item: Ingredient
    rate: Rational = 0
    annotations: str = ''
    def __str__(self):
        return '{}{} {}'.format(self.item, self.annotations, _fmt_rate(self.item, self.rate, _unitForItem(self.item)))
    def __neg__(self):
        return OneWayFlow(self.item, -self.rate, self.annotations)
    def __repr__(self):
        return '<{}>'.format(str(self))

class Flow(NamedTuple):
    item: Ingredient
    rateOut: Rational = 0
    rateIn: Rational = 0
    rateSelf: Rational = 0
    adjusted: bool = False
    underflow: bool = False
    annotation: str = None
    def __bool__(self):
        return self.rateOut == 0 and self.rateIn == 0
    def rate(self):
        return self.rateOut - self.rateIn
    def annotations(self):
        s = ''
        if self.adjusted:
            s += '~'
        if self.annotation is None:
            if self.rateOut > self.rateIn and self.rateIn > 0:
                s += '*'
            elif self.rateOut < self.rateIn and self.rateOut - self.rateSelf > 0:
                s += '!'
        else:
            s += self.annotation
        return s
    def netFlow(self):
        return OneWayFlow(self.item, self.rate(), self.annotations())
    def __str__(self):
        unit = _unitForItem(self.item)
        s = '{}{} '.format(self.item, self.annotations())
        s += '{}'.format(_fmt_rate(self.item, self.rate(), unit))
        if self.rateIn > 0 and self.rateOut > 0:
            s += ' ({}'.format(_fmt_rate(self.item, self.rateOut - self.rateSelf, unit))
            s += ' - {}'.format(_fmt_rate(self.item, self.rateIn - self.rateSelf, unit))
            if self.rateSelf > 0:
                s += ', {} self'.format(_fmt_rate(self.item, self.rateSelf, unit))
            s += ')'
        return s
    def __repr__(self):
        return '<{}>'.format(self.__str__())
    def __mul__(self, factor):
        if factor == 1:
            return self
        else:
            return self.copy(factor = factor)
    __rmul__ = __mul__
    def copy(self, *, factor = 1, adjusted = None, underflow = None, annotation = None):
        return Flow(self.item,
                    self.rateOut * factor,
                    self.rateIn * factor,
                    self.rateSelf * factor,
                    adjusted if adjusted is not None else self.adjusted,
                    underflow if underflow is not None else self.underflow,
                    annotation if annotation is not None else self.annotation)

class FlowsState(OrdEnum):
    OK = 0
    UNDERFLOW = 1
    UNSOLVED = 2
    INVALID = 3
    def ok(self):
        return self.value < FlowsState.UNSOLVED.value

class Flows:
    # self.byItem and self._mainOutput and self.state are expected to be defined
    def flow(self, item):
        if isinstance(item, Ingredient):
            return self.byItem.get(item,Flow(item))
        else:
            raise KeyError(item)
    def rate(self, item):
        return self.flow(item).rate()
    def __iter__(self):
        return iter(self.byItem.values())
    def _showFlow(self, flow):
        return True
    def __repr__(self):
        flowsStr = ' '.join(repr(flow) for flow in self.byItem.values() if self._showFlow(flow))
        if self.state:
            return '<{} {}>'.format(self.state.name, flowsStr)
        else:
            return '<{}>'.format(flowsStr)
    def __str__(self):
        flowsStr = ', '.join(str(flow) for flow in self.byItem.values() if self._showFlow(flow))
        if self.state:
            return '({}) {}'.format(self.state.name, flowsStr)
        else:
            return flowsStr
    def items(self):
        return self.byItem.keys()
    def print(self, out = None, prefix = '  '):
        if out is None:
            out = sys.stdout
        for flow in self.byItem.values():
            if not self._showFlow(flow): continue
            out.write(f'{prefix}{flow}\n')
        if self.state:
            out.write(f'{prefix}{self.state.name}\n')
    __getitem__ = flow
    def __len__(self):
        return len(self.byItem)
    def mainOutput(self):
        return self.flow(self._mainOutput)
    def mul(self,num,markAsAdjusted=False): 
        flows = _MutableFlows(initFrom = self)
        for f in self:
            flows.merge(f,num,markAsAdjusted)
        if not isinstance(self, _MutableFlows):
            flows = self.__class__(flows)
        return flows
    def filter(self, *, items):
        if items is None:
            return self
        flows = _MutableFlows(initFrom = self)
        for item in items:
            try:
                flows.byItem[item] = self.byItem[item]
            except KeyError:
                pass
        if not isinstance(self, _MutableFlows):
            flows = self.__class__(flows)
        return flows
    def __mul__(self,num):
        return self.mul(num)
    def __rmul__(self,num):
        return self.mul(num)
    def __eq__(self, other):
        if not isinstance(other, Flows):
            return False
        return self.byItem == other.byItem and self._mainOutput == other._mainOutput and self.state == other.state

class _MutableFlows(Flows):
    def __init__(self, *, initFrom = None):
        self.byItem = {}
        if initFrom is not None:
            self._mainOutput = initFrom.mainOutput
            self.state = initFrom.state
        else:
            self._mainOutput = None
            self.state = FlowsState.OK
    def _merge(self, flow, num, markAsAdjusted, singleFacility):
        singleFacility = False # fixme: hack
        if flow.item in self.byItem:
            f = self.byItem[flow.item]
            rateOut = f.rateOut + flow.rateOut*num
            rateIn = f.rateIn + flow.rateIn*num
            if singleFacility:
                rateSelf = rateOut if rateOut < rateIn and rateOut > 0 else 0
            else:
                rateSelf = f.rateSelf + flow.rateSelf*num
            adjusted = markAsAdjusted or f.adjusted or flow.adjusted
            underflow = f.underflow or flow.underflow
            self.byItem[flow.item] = Flow(flow.item,rateOut,rateIn,rateSelf,adjusted,underflow)
        else:
            self.byItem[flow.item] = num * flow
    def addFlow(self, item, **kws):
        self._merge(Flow(item, **kws), 1, False, True)
    def merge(self,flow,num=1,markAsAdjusted = False):
        self._merge(flow, num, markAsAdjusted, False)
    def reorder(self):
        def sort(flows):
            return sorted(flows, key=lambda flow: (flow.item.order, flow.item.name, flow.item))
        flows = chain(
            sort(flow for flow in self if flow.rateOut != 0 and flow.rateIn == 0),
            sort(flow for flow in self if flow.rateOut != 0 and flow.rateIn != 0),
            sort(flow for flow in self if flow.rateOut == 0 and flow.rateIn == 0),
            sort(flow for flow in self if flow.rateOut == 0 and flow.rateIn != 0),
        )
        self.byItem = {}
        for f in flows:
            self.byItem[f.item] = f

class SimpleFlows(Flows):
    def __init__(self, mutableFlows):
        self.byItem = mutableFlows.byItem
        self._mainOutput = mutableFlows._mainOutput
        self.state = mutableFlows.state
    def output(self, item):
        return self.flow(item).rateOut
    def input(self, item):
        return self.flow(item).rateIn
    def outputs(self):
        return [OneWayFlow(flow.item, flow.rateOut) for flow in self if flow.rateOut > 0]
    def inputs(self):
        return [OneWayFlow(flow.item, flow.rateIn) for flow in self if flow.rateIn > 0]

class NetFlows(Flows):
   def __init__(self, mutableFlows):
       self.byItem = mutableFlows.byItem
       self._mainOutput = mutableFlows._mainOutput
       self.state = mutableFlows.state
   def output(self, item):
       rate = self.flow(item).rate()
       if rate <= 0: return 0
       return rate
   def input(self, item):
       rate = self.flow(item).rate()
       if rate >= 0: return 0
       return -rate
   def outputs(self):
       return [flow.netFlow() for flow in self if flow.rate() > 0]
   def inputs(self):
       return [-flow.netFlow() for flow in self if flow.rate() < 0]
   def internal(self):
       return [f for f in self if f.rateIn > 0 and f.rateOut > 0]
   def surplus(self):
       return [f for f in self if '*' in f.annotations()]
   def lackof(self):
       return [f for f in self if '!' in f.annotations()]

class EffectBase(NamedTuple):
    speed: Rational = 0
    productivity: Rational = 0
    consumption: Rational = 0 # energy used 
    pollution: Rational = 0

    def __str__(self):
        parts = []
        if self.speed != 0:
            parts.append('{:+.0%} speed'.format(self.speed))
        if self.productivity != 0:
            parts.append('{:+.0%} prod.'.format(self.productivity))
        if self.consumption != 0: 
            parts.append('{:+.0%} energy'.format(self.consumption))
        if self.pollution != 0:
            parts.append('{:+.0%} pollution'.format(self.pollution))
        return ' '.join(parts)

    def __repr__(self):
        return f'<{type(self).__name__}: {str(self)}>'

    def __bool__(self):
        return self.speed != 0 or self.productivity != 0 or self.consumption != 0 or self.pollution != 0

    def __add__(self, other):
        if type(self) is not type(other):
            raise TypeError
        return EffectBase.__new__(type(self),
                                  self.speed + other.speed,
                                  self.productivity + other.productivity,
                                  self.consumption + other.consumption,
                                  self.pollution + other.pollution)

    def __mul__(self, n):
        return EffectBase.__new__(type(self),
                                  self.speed*n,
                                  self.productivity*n,
                                  self.consumption*n,
                                  self.pollution*n)

    def __rmul__(self, n):
        return EffectBase.__new__(type(self),
                                  self.speed*n,
                                  self.productivity*n,
                                  self.consumption*n,
                                  self.pollution*n)

    def __truediv__(self, n):
        return EffectBase.__new__(type(self),
                                  div(self.speed, n),
                                  div(self.productivity, n),
                                  div(self.consumption, n),
                                  div(self.pollution, n))

class Effect(EffectBase):
    pass

class Bonus(EffectBase):
    def __new__(cls, *args, **kwargs):
        if len(args) <= 1 and len(kwargs) == 0:
            if len(args) == 0:
                return EffectBase.__new__(cls)
            other = args[0]
            if type(other) is Bonus:
                return other
            elif type(other) is Effect:
                consumption =  other.consumption if other.consumption > frac(-4, 5) else frac(-4, 5)
                return EffectBase.__new__(cls, other.speed, other.productivity, consumption, other.pollution)
        else:
            return EffectBase.__new__(cls, *args, **kwargs)

class Rcp(Uniq,Immutable):
    """A proxy class for a recipe.  The recipe used depends on the setting of `config.mode`."""
    __slots__ = ('name')
    def __new__(self, name):
        return rcp.byName[name]
    def inst(self, mode = None):
        from . import config
        if mode is None:
            mode = config.mode.get()
        return rcpinst.byName[mode][self.name]

    def __hash__(self):
        return hash(self.name)
    
    @property
    def madeIn(self):
        return self.inst().madeIn
    @property
    def inputs(self):
        return self.inst().inputs
    @property
    def outputs(self):
        return self.inst().outputs
    @property
    def time(self):
        return self.inst().time
    @property
    def order(self):
        return self.inst().order

    def str(self):
        return self.inst().str()

    def produce(self, **kwargs):
        return self.inst().produce(**kwargs)

    def __call__(self, **kwargs):
        return self.inst().__call__(**kwargs)

    def __repr__(self):
        from ._helper import toPythonName
        pythonName = toPythonName(self.name)
        return f'rcp.{pythonName}'

    def __str__(self):
        return f'recipe {self.name}'

class RecipeComponent(NamedTuple):
    num: Rational
    item: Item
    def __str__(self):
        return '{:n} {}'.format(float(self.num), self.item)

class Recipe(Uniq,Immutable):
    """A recipe to produce something.
    
    In most cases it should not be used directly.  Instead use `Rcp` to
    account for the fact that some recipes are different depending if the
    game mode is set to 'normal' or 'expensive'.

    """
    __slots__ = ('name', 'madeIn', 'inputs', 'outputs', 'time', 'order')
    def __init__(self, name, madeIn, inputs, outputs, time, order):
        object.__setattr__(self, 'name', name)
        object.__setattr__(self, 'madeIn', madeIn)
        object.__setattr__(self, 'inputs', tuple(inputs))
        object.__setattr__(self, 'outputs', tuple(outputs))
        object.__setattr__(self, 'time', time)
        object.__setattr__(self, 'order', order)
    def __eq__(self, other):
        if isinstance(other, Rcp):
            return self.name == other.name
        else:
            return object.__eq__(self, other)
    def __ne__(self, other):
        if isinstance(other, Rcp):
            return self.name != other.name
        else:
            return object.__ne__(self, other)
    def __hash__(self):
        return hash(self.name)
    def __repr__(self):
        from ._helper import toPythonName
        pythonName = toPythonName(self.name)
        if getattr(rcpinst, pythonName, None) is self:
            return 'rcpinst.' + pythonName
        elif getattr(rcpinst.expensive, pythonName, None) is self:
            return 'rcpinst.expensive.' + pythonName
        else:
            return '<{}: {}>'.format(type(self).__name__, self.name)
    def str(self):
        return '{} <={:n}s= {}'.format(
            ', '.join(map(str, self.outputs)),
            float(self.time),
            ', '.join(map(str, self.inputs)),
        )
    def _jsonObj(self, customRecipes, **kwards):
        from .jsonconv import _jsonObj
        if rcpinst.byName[Mode.NORMAL].get(self.name, None) is self:
            return self.name
        elif rcpinst.byName[Mode.EXPENSIVE].get(self.name, None) is self:
            return f'{self.name} expensive'
        elif customRecipes.get(self.name, None) is self:
            return f'{self.name} custom'
        if self.name in customRecipes:
            raise ValueError(f'recipe name conflict: {self.name}')
        customRecipes[self.name] = self
        return f'{self.name} custom'
    def eqv(self, other):
        """Returns true if other is equivalent to self and not just the same object."""
        return (type(self) == type(other)
                and self.name == other.name
                and self.madeIn == other.madeIn
                and self.inputs == other.inputs
                and self.outputs == other.outputs
                and self.time == other.time
                and self.order == other.order)
    def produce(self, machinePrefs = Default, fuel = None, modules = (), beacons = ()):
        from collections import deque
        from . import config
        if machinePrefs is Default:
            machinePrefs = config.machinePrefs.get()
        candidates = deque()
        for m in machinePrefs:
            if isinstance(m, self.madeIn) and (m.recipe is None or m.recipe.name == self.name):
                candidates.append(copy(m))
        if not candidates:
            candidates.append(self.madeIn())
        while candidates:
            m = candidates.popleft()
            try:
                m.recipe = self
                break
            except InvalidModulesError as err:
                invalidModules = err.invalid
        else:
            m.modules = tuple(m for m in m.modules if m not in invalidModules)
            m.recipe = self
        if modules is not Default:
            m.modules = modules
        if beacons is not Default:
            m.beacons = beacons
        if hasattr(m, 'fuel'):
            if fuel is Default:
                m.fuel = config.defaultFuel.get()
            else:
                m.fuel = fuel
        return m
    def __call__(self, machinePrefs = Default, fuel = Default, modules = Default, beacons = Default):
        return self.produce(machinePrefs = machinePrefs, fuel = fuel, modules = modules, beacons = beacons)

class InvalidModulesError(ValueError):
    def __init__(self, m):
        self.invalid = m
    pass

IGNORE = -100



from __future__ import annotations
from typing import NamedTuple as _NamedTuple

from .core import Electricity as _Electricity, Item as _Item
from .fracs import frac as _frac

class Unit(_NamedTuple):
    abbr: str
    sep: str
    conv: Rational

def wagonsPerUnit(item, rate, conv):
    from .fracs import div
    try:
        wagonCapacity = item.stackSize*40
    except AttributeError:
        # assume item is a fluid
        wagonCapacity = 25000
    return div(conv * rate, wagonCapacity)

def wagonsPerMinute(*args):
    if len(args) == 1:
        return wagonsPerUnit(flow.item, flow.rate(), 60)
    elif len(args) == 2:
        return wagonsPerUnit(args[0], args[1], 60)
    else:
        raise ValueError('usage: wagonsPerMinute(flow) or waginsPerMinute(item, rate)')

UNIT_SECONDS        = Unit('s', '/',  1)
UNIT_MINUTES        = Unit('m', '/',  60)
UNIT_HOURS          = Unit('h',  '/', 3600)
UNIT_EXPRESS_BELTS  = Unit('eb', ' ', _frac(1, 45))
UNIT_FAST_BELTS     = Unit('fb', ' ', _frac(1, 30))
UNIT_TRANSFER_BELTS = Unit('tb', ' ', _frac(1, 15))
UNIT_MEGAWATT       = Unit('MW', ' ', 1)
UNIT_WAGONS_PER_MIN = Unit('w/m', ' ', wagonsPerMinute)

DU_SECONDS =  ((_Electricity, UNIT_MEGAWATT), (None, UNIT_SECONDS))
DU_MINUTES =  ((_Electricity, UNIT_MEGAWATT), (None, UNIT_MINUTES))
DU_HOURS   =  ((_Electricity, UNIT_MEGAWATT), (None, UNIT_HOURS))
DU_EXPRESS_BELTS  = ((_Item, UNIT_EXPRESS_BELTS),  (_Electricity, UNIT_MEGAWATT), (None, UNIT_SECONDS))
DU_FAST_BELTS     = ((_Item, UNIT_FAST_BELTS),     (_Electricity, UNIT_MEGAWATT), (None, UNIT_SECONDS))
DU_TRANSFER_BELTS = ((_Item, UNIT_TRANSFER_BELTS), (_Electricity, UNIT_MEGAWATT), (None, UNIT_SECONDS))
DU_WAGONS_PER_MIN = ((_Electricity, UNIT_MEGAWATT), (None, UNIT_WAGONS_PER_MIN))

__all__ = [sym for sym in globals() if not sym.startswith('_') and sym not in ('annotations')]



"""Solver intervals.

Example::

  >>> from factoriocalc import *
  >>> config.machinePrefs.set(MP_MAX_PROD.withSpeedBeacons({AssemblingMachine3:8, ChemicalPlant:8, OilRefinery:12}))
  >>> rocketFuel = UnboundedBox(1*rcp.advanced_oil_processing() 
                                + 1*rcp.heavy_oil_cracking() 
                                + 1*rcp.light_oil_cracking()
                                + 1*rcp.solid_fuel_from_light_oil() 
                                + 1*rcp.solid_fuel_from_petroleum_gas()
                                + 1*rcp.rocket_fuel(),
                                outputs = [itm.rocket_fuel@6])
  >>> solver = rocketFuel.solver()
  >>> solver.reset()
  <SolveRes.UNSOLVED: -2>
  >>> solver.print()
  +113.258*light-oil-cracking-m -45.5*solid-fuel-from-petroleum-gas-m +18.0360*solid-fuel-from-light-oil-m +1.45345*rocket-fuel-m = 0 (140251354372976-petroleum-gas)
  +2.9575*solid-fuel-from-light-oil-m +2.9575*solid-fuel-from-petroleum-gas-m -1.83333*rocket-fuel-m = 0 (140251354372976-solid-fuel)
  +0.256667*rocket-fuel-m = 6 (140251354372976-rocket-fuel)
  advanced-oil-processing-m = +0.577677*light-oil-cracking-m +0.192559*solid-fuel-from-light-oil-m +0.0155176*rocket-fuel-m
  heavy-oil-cracking-m = +0.270270*light-oil-cracking-m +0.0900901*solid-fuel-from-light-oil-m +0.00726001*rocket-fuel-m
  max (inputs): -200.209*light-oil-cracking-m -43.9865*solid-fuel-from-light-oil-m -3.54470*rocket-fuel-m
    aux (water): -124.534*light-oil-cracking-m -18.7613*solid-fuel-from-light-oil-m -1.51190*rocket-fuel-m
    aux (crude-oil): -75.6757*light-oil-cracking-m -25.2252*solid-fuel-from-light-oil-m -2.03280*rocket-fuel-m
  >>> solver.tableau.print()
        -0-    -1-    -2-    -3-   -4-  -5-  -6- 
      light- solid- solid- rocket  a0   a1   p2  |    rhs    
   0:  0.635 -0.255  0.101  0.008  *1*   0    0  |     0     | -4- a0
   1:    0    0.381  0.381 -0.236   0   *1*   0  |     0     | -5- a1
   2:    0      0      0      1     0    0   *1* |  23.3766  | -6- p2

   3: -0.635 -0.126 -0.482  0.228   0    0    0  |     0     | max
   4:    0      0      0     -1     0    0    0  | -23.3766  | max
  >>> solver.tableau.solve(zero=True)
  <SolveRes.OPTIMAL: 1>
  >>> solver.tableau.solve(zero=True)
  <SolveRes.OPTIMAL: 1>
  >>> solver.tableau.addPendingObjective()
  >>> solver.tableau.print()
        -0-    -1-    -2-    -3-  
      light- solid- solid- rocket |    rhs    
   0:   *1*     0    0.560    0   |  5.52158  | -0- light-oil-cracking-m
   1:    0     *1*     1      0   |  14.4910  | -1- solid-fuel-from-petroleum-gas-m
   2:    0      0      0     *1*  |  23.3766  | -3- rocket-fuel-m

   3:    0      0   -68.32    0   | -1188.34  | max (inputs)
   4:    0      0   -51.10    0   | -722.966  | aux (water)
   5:    0      0   -17.22    0   | -465.369  | aux (crude-oil)
  >>> solver.tableau.solve()
  <SolveRes.UNIQUE: 2>
  >>> solver.tableau.print()
        -0-    -1-    -2-    -3-  
      light- solid- solid- rocket |    rhs    
   0:  1.782    0     *1*     0   |  9.84266  | -2- solid-fuel-from-light-oil-m
   1: -1.782   *1*     0      0   |  4.64834  | -1- solid-fuel-from-petroleum-gas-m
   2:    0      0      0     *1*  |  23.3766  | -3- rocket-fuel-m

   3:  121.8    0      0      0   | -515.807  | max (inputs)
   4:  91.09    0      0      0   | -220.004  | aux (water)
   5:  30.70    0      0      0   | -295.803  | aux (crude-oil)
  >>> solver.tableau.solution().print()
  Solution:
    light-oil-cracking-m = 0
    solid-fuel-from-petroleum-gas-m = 4.64833996
    solid-fuel-from-light-oil-m = 9.84266354
    rocket-fuel-m = 23.3766234
  Other:

"""

from __future__ import annotations
from enum import Enum,Flag
import operator
from random import Random
from copy import deepcopy
from numbers import Number
from types import SimpleNamespace
from functools import partialmethod
from collections import defaultdict
from collections.abc import Mapping
import io
import os
import itertools
from typing import NamedTuple
from dataclasses import dataclass,field
import sys

from . import itm
from .fracs import Inf,NaN,frac,div,isfinite
from .ordenum import OrdEnum
from .core import *
from .box import *
from .core import _fmt_rate

__all__ = ('Var', 'Term', 'SolveRes', 'Cond', 'EQ', 'GE', 'LE', 
           'LinearEqId', 'Terms', 'LinearEq', 'VarGroup', 'LinearEqSystem',
           'Solver', 'SlackVar', 'ArtificialVar', 'PartialVar', 'OptFun',
           'Tableau', 'Solution')

class Var:
    __slots__ = ('id', 'name', 'max')
    def __init__(self, id, name = None, max = 1):
        if name is None:
            name = id
        if type(name) is not str:
            raise ValueError
        self.id = id
        self.name = name
        self.max = max
    def __eq__(self, other):
        if type(other) is Var:
            return self.id == other.id
        else:
            return self.id == other
    def __hash__(self):
        return hash(self.id)
    def __str__(self):
        return self.name
    def __repr__(self):
        return f'<var: {self.name}>'

class Term(tuple):
    """Term(var, rate)"""
    __slots__ = ()
    var = property(operator.itemgetter(0))
    var.__doc__ = None
    rate = property(operator.itemgetter(1))
    rate.__doc__ = None
    def __new__(cls, *args):
        l = len(args)
        if l == 0:
            args = (None, 0)
        elif l == 1:
            args = (None, args[0])
        return tuple.__new__(cls, args)
    def __str__(self):
        rate = f'{self.rate:.6g}'
        if self.var == None:
            return rate
        else:
            return f'{self.var}*{rate}'
    def __repr__(self):
        return f'Term({self.var!r}, {self.rate!r})'

class SolveRes(OrdEnum):
    """Result of solving a linear equation system.

    The values of the enum are subject to change, but the relative order should be stable.

    """

    #: not yet solved
    UNSOLVED = -2

    #: already solved 
    NOOP = -1

    #: a solution was found, no test for uniqueness was done
    OPTIMAL = 1

    #: all output machines are at there maxium giving the constraints
    UNIQUE = 2

    #: output machines are at there maxium giving the constraints
    #: but there are multiple possible configurations
    MULTI = 3

    #: a solution was found but some machines are not at there maxium output,
    #: to make the solution unique add additional constraints or priorities
    OK = 4

    #: a partial solution was found by removing constraints
    PARTIAL = 5

    #: the system is unbounded, a partial solution was found by setting
    #: unbounded vars to 0
    UNBOUNDED = 6

    # #: no solution was found (unused)
    # ERROR = 9

    # fixme ?: maybe use flags to return more information as a solution can be
    #   both partial and unique
    # SOLVED = 1
    # UNIQUE = 2
    # MULTI  = 4
    # PARTIAL = 8
    # UNBOUNDED = 16
    
    def ok(self):
        return self < SolveRes.PARTIAL
    def notok(self):
        return self >= SolveRes.PARTIAL
    def failed(self):
        return self >= SolveRes.UNBOUNDED

class Cond(NamedTuple):
    order: int
    symbol: str
    fun: Callable[[Rational, Rational], bool]
    def __call__(self, x, y = 0):
        return self.fun(x,y)
    def __repr__(self):
        return '<Cond "{}">'.format(self.symbol)
    def __eq__(self, other):
        return id(self) == id(other)
    def __ne__(self, other):
        return id(self) != id(other)
    def __copy__(self):
        return self
    def __deepcopy__(self, memo):
        return self

EQ = Cond(1, '=', operator.eq)
GE = Cond(2, '>=', operator.ge)
LE = Cond(3, '<=', operator.le)
Cond.NONE = Cond(5, '?', None)

class LinearEqId(NamedTuple):
    boxid: int
    item: Union[Var,Ingredient]
    qualifier: str = ''
    def __str__(self):
        if self.qualifier:
            return f'{self.boxid}-{self.item}-{self.qualifier}'
        else:
            return f'{self.boxid}-{self.item}'
    def __repr__(self):
        return f'<eqid: {self}>'

@dataclass
class Terms(dict):
    def __init__(self, terms = ()):
        if isinstance(terms, Mapping):
            super().__init__(terms)
        else:
            for vals in terms:
                self.__iadd__(vals)

    def __getitem__(self):
        return self.get(other.var, 0)

    def asSeq(self):
        return tuple(Term(var, rate) for var, rate in self.items())

    def add(self, var, rate):
        rate = self.get(var, 0) + rate
        if rate == 0:
            self.pop(var)
        else:
            self[var] = rate

    def merge(self, other, mul = 1):
        if isinstance(other, Term):
            self.add(other.var, mul*other.rate)
            return self
        if isinstance(other, Mapping):
            itr = other.items()
        else:
            itr = iter(other)
        for var, rate in itr:
            self.add(var, mul*rate)
        return self
                
    def apply(self, values = None):
        if values is None:
            values = {}
        res = 0
        for var, rate in self.items():
            res += rate * values.get(var, var.max)
        return res
    
    def sub(self, var, repl):
        factor = self.pop(var, 0)
        if factor == 0:
            return self
        return self.merge(repl, factor)

    def subAll(self, repl):
        vars = [var for var in self.keys() if var in repl]
        for var in vars:
            self.sub(var, repl[var])
        return self

    def neg(self):
        for var, rate in self.items():
            self[var] = -rate
        return self
    
    def __str__(self):
        return fmt_terms(self.items())
    __iadd__ = merge
    __isub__ = partialmethod(merge, mul = -1)
    def __imul__(self, other):
        if not isinstance(other, Number):
            return NotImplemented
        for var in self.keys():
            self[var] *= other
        return self
    def __itruediv__(self, other):
        if not isinstance(other, Number):
            return NotImplemented
        for var, rate in self.items():
            self[var] = div(rate, other)
        return self

def fmt_terms(terms):
    return ' '.join('{}{}'.format('+' if rate == 1 else '-' if rate == -1 else f'{rate:+.6g}*', var) for var, rate in terms)

@dataclass(frozen = True)
class LinearEq:
    # the linear equation is: terms... <cond> rate
    # i.e if cond is EQ then: terms... = rate
    id: LinearEqId
    terms: tuple
    cond: Cond = None
    rate: Rational = None
    def __post_init__(self):
        assert(type(self.terms) is tuple)
    def _replace(self, **changes):
        import dataclasses as dc
        return dc.replace(self, **changes)
    def flow(self, solution, default = Inf):
        rateOut = 0
        rateIn = 0
        for term in self.terms:
            if term.rate > 0:
                rateOut += term.rate*solution.get(term.var, default)
            elif term.rate < 0:
                rateIn -= term.rate*solution.get(term.var, default)
        return Flow(self.id, rateOut = rateOut, rateIn = rateIn)
    def solved(self, solution, default = Inf):
        rate = self.flow(solution, default).rate()
        return self.cond(rate, self.rate) is True
    def __str__(self):
        return f'{fmt_terms(self.terms)} {self.cond.symbol} {self.rate:.6g} ({self.id})'
    def sympify(self):
        from sympy import Number,Symbol,Eq,Ge,Le
        terms = sum(Number(term.rate) * Symbol(term.var.name) for term in self.terms)
        if self.cond is EQ:
            return Eq(terms, Number(self.rate))
        elif self.cond is GE:
            return Ge(terms, Number(self.rate))
        elif self.cond is LE:
            return Le(terms, Number(self.rate))
        else:
            raise ValueError
        
class _Target(NamedTuple):
    cond: Cond
    rate: Rational

@dataclass
class _LinearEq:
    id: Ingredient
    terms: list[Term] = field(default_factory = list)
    target: list[_Target] = field(default_factory = list)

class _Eqs(dict):
    def __missing__(self, id):
        val = self[id] = _LinearEq(id)
        return val

class VarGroup(Group):
    def __init__(self, var):
        super().__init__([])
        self.var = var

class LinearEqSystem:
    def print(self, out = None, prefix=''):
        out = sys.stdout if out is None else out
        for eq in self.eqs.values():
            out.write(f'{prefix}{eq}\n')
        byPrior = defaultdict(list)
        for var, p in self.priorities.items():
            byPrior[p].append(var)
        for p in sorted(byPrior.keys(), reverse = True):
            vars = ', '.join(str(var) for var in byPrior[p])
            out.write(f'{prefix}{vars} (priorities = {p})\n')
        outputs = sorted(self.outputs.keys(),
                         key = lambda item: self.outputPriorities.get(item, 0),
                         reverse = True)
        for item in outputs:
            terms = self.outputs[item]
            priorities = self.outputPriorities.get(item, None)
            if priorities is None:
                out.write(f'{prefix}{item} = {fmt_terms(terms)} (output)\n')
            else:
                out.write(f'{prefix}{item} = {fmt_terms(terms)} (output; priorities = {priorities})\n')
        for item, terms in self.inputs.items():
            out.write(f'{prefix}{item} = {fmt_terms(terms)} (input)\n')
        for item, terms in self.other.items():
            out.write(f'{prefix}{item} = {fmt_terms(terms)} (other)\n')
        

    @classmethod
    def fromBox(cls, box, *, _tally = None):
        byVar = {}
        eqs = _Eqs()
        priorities = {}
        outputPriorities = {}
        inputPriorities = {}

        if isinstance(box, UnboundedBox):
            machines = box.inner.machines
        else:
            machines = box.inner.flatten().machines

        def varInfo(m):
            m = m.machine
            if isinstance(m, BlackBox):
                return ((id(box), id(m)), 'BOX' if m.name is None else m.name)
            elif m.recipe is not None:
                if hasattr(m, 'fuel'):
                    return ((id(box), m.recipe, m.fuel), f'{m.recipe.name}-u-{m.fuel}')
                else:
                    return ((id(box), m.recipe), m.recipe.name)
            else:
                return (None, 'unknown')

        varMax = Inf if isinstance(box, UnboundedBox) else 1

        tally = _tally
        if tally is None:
            tally = defaultdict(dict)
        tally[''][id(box)] = len(tally['']) + 1
            
        for m in machines:
            (key, name) = varInfo(m) 
            tally[name][key] = len(tally[name]) + 1

        inners = []
        for idx, m in enumerate(machines):
            if not isinstance(m.machine, Box):
                continue
            l = cls.fromBox(m.machine, _tally = tally)
            l.box = m.machine
            l *= m.num
            inners.append(l)
            machines[idx] = None
            
        for m in machines:
            if m is None: continue
            (key, name) = varInfo(m)
            suffix = 't' if isfinite(varMax) else 'm'
            if len(tally[name]) > 1:
                num = tally[name][key]
                name = f'{name}-{num-1}-{suffix}'
            else:
                name = f'{name}-{suffix}'
            var = Var(key,name,varMax)
            if var not in byVar:
                byVar[var] = VarGroup(var)
            byVar[var].machines.append(m)

        for var, grp in byVar.items():
            if varMax == Inf:
                if len(grp) != 1 or not isinstance(grp[0], Mul):
                    raise ValueError('UnboundedBox is not in proper form')
                grp.flows = grp[0].machine.flows(throttle = 1)
            else:
                grp.flows = grp.flows(throttle = 1)
            for flow in grp.flows:
                #assert(flow.rate() != 0)
                if flow.rate() == 0:
                    continue
                term = Term(var, flow.rate())
                eqs[flow.item].terms.append(term)

        for l in inners:
            for item, terms in l.external():
                eqs[item].terms += terms
            byVar.update(l.byVar)                
        
        for (item,rate) in box.outputs.items():
            row = eqs[item]
            if rate is None:
                if any(term.rate < 0 for term in row.terms):
                    row.target.append(_Target(GE, 0))
            else:
                row.target.append(_Target(EQ, rate))
            if rate != 0:
                row.target.append(_Target(Cond.NONE, 1))
        for (item,rate) in box.inputs.items():
            if item in box.outputs: # if in both input and output then value in unconstrained
                eqs[item].target = [_Target(Cond.NONE, 0)]
                continue
            row = eqs[item]
            if rate is None:
                if any(term.rate > 0 for term in row.terms):
                    row.target.append(_Target(LE, 0))
            else:
                row.target.append(_Target(EQ, rate))
            if rate != 0:
                row.target.append(_Target(Cond.NONE, -1))
        for row in eqs.values():
            if not row.target and row.id is not itm.electricity:
                row.target = [_Target(EQ, 0)]
        for item, rate in box.constraints.items():
            if type(rate) is tuple:
                continue # deal with it later
            row = eqs[item]
            row.target.append(_Target(GE, rate))
            if rate < 0 and any(term.rate > 0 for term in row.terms):
                row.target.append(_Target(LE, 0))

        ###

        for (var, grp) in byVar.items():
            for flow in grp.flows:
                if flow.item in box.priorities and flow.item in box.outputs:
                    outputPriorities[flow.item] = box.priorities[flow.item]
                elif flow.item in box.priorities and flow.item in box.inputs:
                    inputPriorities[flow.item] = box.priorities[flow.item]
                elif flow.item in box.outputs and box.outputs[flow.item] is None and flow.item != itm.empty_barrel:
                    outputPriorities[flow.item] = 0

        byId = {}
        for var in byVar:
            byId[var.id] = var
            
        for key0, prior in box.priorities.items():
            key = (id(box), key0)
            if key in byId:
                priorities[byId[key]] = prior

        # fixme: still needed?
        toDel = [var for var, priority in priorities.items() if priority <= IGNORE]
        for var in toDel:
            del priorities[var]

        ###

        origEqs = eqs
        eqs = {}
        external = {1: {}, -1: {}, 0: {}, None: {}}
        for eq in sorted(origEqs.values(), key = lambda v: v.target):
          if not eq.terms:
              continue
          for target in eq.target:
              if target.cond is Cond.NONE:
                  external[target.rate][eq.id] = tuple(eq.terms)
                  external[None][eq.id] = tuple(eq.terms)
                  continue
              q = ''
              if target.cond is GE:
                  q = 'ge'
              elif target.cond is LE:
                  q = 'le'
              eqId = LinearEqId(id(box), eq.id, q)
              eqs[eqId] = LinearEq(eqId, tuple(eq.terms), target.cond, target.rate)

        for item1, rate in box.constraints.items():
            if type(rate) is not tuple:
                continue
            mul, item2 = rate
            eqId = LinearEqId(id(box), item1, 'c')
            eqs[eqId] = LinearEq(eqId,
                                 (*external[None][item1], *(Term(v, -r*mul) for v, r in external[None][item2])),
                                 EQ, 0)
            print(eqs[eqId])
              
        #sortedEqs = sorted(eqs.values(), key = lambda v: v.target)
        #eqs = {row.id: row for row in sortedEqs}

        ###
        
        for l in inners:
            for eqId, eq in l.eqs.items():
                if eqId.qualifier == '':
                    eqId = eqId._replace(qualifier = 'b{}'.format(str(tally[''][id(l.box)] - 1)))
                eqs[eqId] = eq._replace(id = eqId)

        obj = LinearEqSystem()
        obj.eqs = eqs
        obj.byVar = byVar
        obj.outputs = external[1]
        obj.inputs = external[-1]
        obj.other = external[0]
        obj.priorities = priorities
        obj.outputPriorities = outputPriorities
        obj.inputPriorities = inputPriorities
        return obj

    def __imul__(self, num):
        if num == 1:
            return self
        for eqId, eq in self.eqs.items():
            self.eqs[eqId] = eq._replace(terms = tuple(Term(var, rate*num) for var, rate in eq.terms),
                                         rate = eq.rate * num)
        for io in (self.outputs, self.inputs, self.other):
            for item, terms in io.items():
                io[item] = tuple(Term(var, rate*num) for var, rate in terms)
        return self

    def external(self):
        return [*self.outputs.items(), *self.inputs.items(), *self.other.items()]

    def sympify(self):
        from sympy import Symbol
        return {'eqs': [eq.sympify() for eq in self.eqs.values()],
                'vars': [Symbol(var.name) for var in self.byVar.keys()]}

    def solver(self):
        return Solver(self)

@dataclass
class _Equation:
    id: LinearEqId
    terms: Terms
    cond: Cond
    rhs: Term
    def __str__(self):
        return f'{self.terms} {self.cond.symbol} {self.rhs} ({self.id})'

def _simplify(leq):
    eqs = []
    for eq in leq.eqs.values():
        eqs.append(_Equation(eq.id, Terms(eq.terms), eq.cond, Term(eq.rate)))
    ZERO = 0
    while True:
        candidates = {}
        for eq in eqs:
            if eq.rhs.rate != 0: continue
            posVars = [var for var, rate in eq.terms.items() if rate > 0]
            allPositive = len(posVars) == len(eq.terms) and (eq.cond is EQ or eq.cond is LE)
            allNegative = len(posVars) == 0             and (eq.cond is EQ or eq.cond is GE)
            if allPositive or allNegative:
                for var in eq.terms.keys():
                    candidates[var] = ZERO
            elif eq.cond is EQ and len(posVars) == 1:
                var = posVars[0]
                if var not in candidates:
                    candidates[var] = eq
                elif candidates[var] is not ZERO:
                    candidates[var] = None
        for var in list(candidates.keys()):
            if candidates[var] is None:
                del candidates[var]
        if not candidates:
            break
        for var, eq in candidates.items():
            if eq is ZERO:
                eqs.append(_Equation(None, Terms(), EQ, Term(var,1)))
                terms = ()
            elif eq.terms:
                factor = -eq.terms.pop(var)
                eq.terms /= factor 
                eq.rhs = Term(var, 1)
                terms = eq.terms
            else: # empty eq -- fixme: is this right, why can this happen
                eq.rhs = Term(var, 1)
                terms = eq.terms
            for eq2 in eqs:
                eq2.terms.sub(var, terms)
    return eqs

class Solver:
    def __init__(self, leqs):
        self.leqs = leqs
        self.eqs = None

    def reset(self, simplify = True):
        """Reset the state and does preliminary work for solving.

        Called automatically by solve().  But can be called manually to get
        access to the underlying tableau.

        """
        self.subs = {}
        self._result = SolveRes.UNSOLVED
        self._invalidEqs = []
        if simplify:
            self.eqs = []
            eqs = _simplify(self.leqs)
            for eq in eqs:
                if eq.rhs.var is None:
                    if eq.terms:
                        if eq.cond is GE and eq.rhs.rate == 0 and all(rate >= 0 for rate in eq.terms.values()):
                            # always true
                            pass
                        else:
                            self.eqs.append(LinearEq(eq.id, eq.terms.asSeq(), eq.cond, eq.rhs.rate))
                    elif not eq.cond(0, eq.rhs.rate):
                        self._invalidEqs.append(eq)
                        self._result |= SolveRes.PARTIAL
                    else:
                        pass
                else:
                    assert(eq.cond is EQ)
                    assert(eq.rhs.rate == 1)
                    var = eq.rhs.var
                    self.subs[var] = eq.terms
                    if var.max != Inf and eq.terms.apply() > var.max:
                        self.eqs.append(LinearEq(eq.id, eq.terms.asSeq(), LE, var.max))
        else:
            self.eqs = self.leqs.eqs.values()

        priorities = defaultdict(dict)
        for item, terms in self.leqs.outputs.items():
            try:
                p = self.leqs.outputPriorities[item]
            except KeyError:
                continue
            if p <= IGNORE:
                continue
            for term in terms:
                if term.rate > 0:
                    priorities[p][term.var] = 1
                elif term.rate < 0 and term.var not in priorities[p]:
                    priorities[p][term.var] = -1

        for var, p in self.leqs.priorities.items():
            priorities[p][var] = 1

        objectives = []
        for p in sorted(priorities.keys(), reverse=True):
            toMax, toMin = [], []
            for var, direction in priorities[p].items():
                if direction > 0:
                    toMax.append(var)
                else:
                    toMin.append(var)
            if toMax:
                # maximize all vars
                objtv = [OptFun(defaultdict(dict), f'{p}-max')]
                for var in toMax:
                    objtv[0].terms[var] = 1
                    objtv.append(OptFun({var: 1}, var))
                objectives.append(objtv)
            if toMin:
                # minimize all vars
                objtv = [OptFun(defaultdict(dict), f'{p}-min')]
                for var in toMin:
                    objtv[0].terms[var] = -1
                    objtv.append(OptFun({var: -1}, var))
                objectives.append(objtv)

        # minimize the amount of extra output
        alsoOutput = [item for item, p in self.leqs.outputPriorities.items() if p <= IGNORE]
        if alsoOutput:
           also = [OptFun(Terms(self.leqs.outputs[item] for item in alsoOutput).subAll(self.subs).neg(), 'also-output')]
           for item in alsoOutput:
               also.append(OptFun(Terms(self.leqs.outputs[item]).subAll(self.subs).neg(), item))
           objectives.append(also)

        ignoredInputs, inputs = {}, {}
        for item, terms in self.leqs.inputs.items():
            if self.leqs.inputPriorities.get(item, Inf) <= IGNORE:
                ignoredInputs[item] = terms
            else:
                inputs[item] = terms

        # maximize the machines that produce an ignored input
        if ignoredInputs:
            also = [OptFun(defaultdict(dict), 'ignored-inputs')]
            for item, terms in ignoredInputs.items():
                also.append(OptFun(Terms(terms).subAll(self.subs), item))
                for term in terms:
                    if term.rate > 0:
                        also[0].terms[term.var] = 1
            objectives.append(also)

        # minimize the amount of input by maximizing the total input flow as
        # input valaues are always negative
        if inputs:
            also = [OptFun(Terms(inputs.values()).subAll(self.subs), 'inputs')]
            for item, terms in inputs.items():
                also.append(OptFun(Terms(terms).subAll(self.subs), item))
            objectives.append(also)

        for obj in objectives:
            if len(obj[0]) == 1:
                obj[:] = [obj[0]]

        self.objectives = objectives
        
        self.tableau = Tableau(self.eqs, self.objectives)
        return self._result

    prep = reset

    def print(self, out = None):
        out = sys.stdout if out is None else out
        if self.eqs is None:
            out.write('<<uninitialized>>\n')
            return
        for eq in self.eqs:
            out.write(f'{eq}\n')
        for var, terms in self.subs.items():
            out.write(f'{var} = {terms}\n')
        for objective in self.objectives:
            out.write(f'max ({objective[0].note}): {fmt_terms(objective[0].terms.items())}\n')
            for i in range(1,len(objective)):
                out.write(f'  aux ({objective[i].note}): {fmt_terms(objective[i].terms.items())}\n')

    def solve(self):
        """Solve the linear equation system."""
        if self.eqs is None or self.subs is None:
            self.reset()
        self._result |= self.tableau.solveAll()

    def _sanity(self):
        solution, other = self.tableau.solution()
        for eq in self.eqs:
            assert(eq.solved(solution))

    def solution(self):
        """Return the solution to the solved linear equation.

        Return a tuple pair. The first component is a SolveRes.  The second
        component is a `dict` with the solution to the system.
        """
        solution, other = self.tableau.solution()
        res = self._result
        if res.ok() and other:
            res = SolveRes.PARTIAL
        for var, terms in self.subs.items():
            rate = 0
            for v, r in terms.items():
                try:
                    rate += r * solution[v]
                except KeyError:
                    if isfinite(v.max):
                        rate += r * v.max
                    else:
                        res |= SolveRes.UNBOUNDED
            solution[var] = rate
        return res, solution

    def apply(self):
        """Apply the solution the the underlying box.

        Returns the same value as solution().
        """
        res, solution = self.solution()
        if res.failed():
            return res, solution
        for var,grp in self.leqs.byVar.items():
            if var.max == 1:
                grp.setThrottle(solution.get(var, 1))
            elif var.max == Inf:
                if len(grp) != 1 or not isinstance(grp[0], Mul):
                    raise ValueError(f'can not apply to {var}')
                grp[0].num = solution.get(var, 1)
            else:
                raise ValueError(f'{var}: unrecognized var type')
        return res, solution
        

@dataclass(frozen=True)
class SlackVar:
    __slots__ = ('rowIdx')
    rowIdx: int
    def __str__(self):
        return f's{self.rowIdx}'

@dataclass(frozen=True)
class ArtificialVar:
    __slots__ = ('rowIdx')
    rowIdx: int
    def __str__(self):
        return f'a{self.rowIdx}'

@dataclass(frozen=True)
class PartialVar(ArtificialVar):
    __slots__ = ()
    def __str__(self):
        return f'p{self.rowIdx}'

class OptFun(NamedTuple):
    terms: Any
    note: Any = None

@dataclass
class _OptFun:
    label: str # 'max' or 'aux'
    terms: Any
    note: Any = None
    nonOptimal: bool = False

class Tableau:
    def __init__(self, eqs, objectives):
        cols = []
        eqs0 = eqs
        eqs = {}
        
        def addEq(terms, minv, maxv):
            #assert(minv <= maxv)
            factor = 0
            for var, rate in terms:
                factor += abs(rate)
                if var not in cols:
                    cols.append(var)
            assert(factor != 0)
            if minv <= 0 and maxv == Inf:
                factor *= -1
            terms = frozenset(Term(var, div(rate, factor)) for var, rate in terms)
            minv = div(minv, factor)
            maxv = div(maxv, factor)
            if factor < 0:
                minv, maxv = maxv, minv
            orig = eqs.get(terms, None)
            if orig is None:
                combined = (minv, maxv)
            else:
                combined = (max(orig[0], minv),
                            min(orig[1], maxv))
            eqs[terms] = combined
        
        for eq in eqs0:
            if eq.cond is EQ:
                addEq(eq.terms, eq.rate, eq.rate)
            elif eq.cond is LE:
                addEq(eq.terms, -Inf, eq.rate)
            elif eq.cond is GE:
                addEq(eq.terms, eq.rate, Inf)
            else:
                raise ValueError
            
        for var in cols:
            if var.max != Inf:
                t = (Term(var, 1),)
                addEq(t, -Inf, var.max)
        
        eql = []
        for rowIdx, (terms, (minv,maxv)) in enumerate(eqs.items()):
            if minv == maxv:
                eql.append((terms, EQ, minv))
                continue
            #assert(minv < maxv)
            if minv > -Inf:
                terms0 = frozenset(Term(var, -rate) for var, rate in terms)
                max0 = -minv
                eql.append((terms0, LE, max0))
            if maxv < Inf:
                eql.append((terms, LE, maxv))

        self.cols = cols
                
        slackVarsOffset = len(cols)
        slackVars = {}
        for rowIdx, (_, cond, _) in enumerate(eql):
            if cond is LE:
                slackVars[rowIdx] = slackVarsOffset + len(slackVars)
                cols.append(SlackVar(rowIdx))
        
        artificialVarsOffset = len(cols)
        artificialVars = {}
        for rowIdx, (_, cond, rhs) in enumerate(eql):
            if rhs < 0 or cond is EQ:
                artificialVars[rowIdx] = artificialVarsOffset + len(artificialVars)
                if rhs == 0:
                    cols.append(ArtificialVar(rowIdx))
                else:
                    cols.append(PartialVar(rowIdx))

        self.zeroed = [False]*len(cols)
        self.colsByVar = colsByVar = {var: idx for idx, var in enumerate(cols)}
        self._unboundedVars = []

        self.tableau = []
        self.basic = []
        self._cleared = None
        for rowIdx, (terms, cond, rhs) in enumerate(eql):
            sign = 1 if rhs >= 0 else -1
            row = [0] * (len(cols) + 1)
            for term in terms:
                row[colsByVar[term.var]] = sign*term.rate
            if rowIdx in slackVars:
                assert(cond is LE)
                colIdx = slackVars[rowIdx]
                row[colIdx] = sign
            if rowIdx in artificialVars:
                colIdx = artificialVars[rowIdx]
                row[colIdx] = 1
            # a row must have an slackVars or artificialVars so colIdx must be defined
            self.basic.append(self.cols[colIdx])
            row[-1] = sign*rhs
            self.tableau.append(row)
        self.optFunBegin = len(self.tableau)
        self.optFunInfo = [None]*len(self.tableau)
        self.allOptFuns = []
        if artificialVars:
            objective1 = {}
            objective2 = {}
            for rowIdx, colIdx in artificialVars.items():
                if type(self.cols[colIdx]) is not PartialVar:
                    objective1[colIdx] = -1
                else:
                    objective2[colIdx] = -1
            if objective1:
                self.addObjective(objective1)
            if objective2:
                self.addObjective(objective2)

        self.pendingObjectives = objectives
        
    def print(self, out = None, ratioCol = None):
        if out is None:
            out = sys.stdout
        out.write('    ')
        for idx, var in enumerate(self.cols):
            s = f'-{idx}-'
            if type(var) is Var:
                out.write(f'{s:^7}')
            else:
                out.write(f'{s:^5}')
        out.write('\n')
        out.write('    ')
        for var in self.cols:
            if type(var) is Var:
                out.write(f'{var!s:^7.6}')
            else:
                out.write(f'{var!s:^5.4}')
        out.write('|    rhs    ')
        if ratioCol is not None:
            out.write(f'|   /{ratioCol}')
        out.write('\n')
        for rowIdx, row in enumerate(self.tableau):
            if rowIdx == self.optFunBegin:
                out.write('\n')
            rhs = row[-1]
            out.write(f'{rowIdx:2}: ')
            basic = self.basic[rowIdx] if rowIdx < len(self.basic) else None
            for var, rate in zip(self.cols, row):
                if type(var) is Var:
                    width = 7
                else:
                    width = 5
                if var == basic:
                    s = f'*{rate:.{width-3}g}*'
                else:
                    s = f'{rate: .{width-1}g}'
                out.write(f'{s:^{width}.{width-1}}')
            s = f'{rhs: .6g}'
            out.write(f'|{s:^11.11}')
            if rowIdx < self.optFunBegin:
                num = self.colsByVar[self.basic[rowIdx]]
                out.write(f'| -{num}- {basic}')
            else:
                out.write(f'| {self.optFunInfo[rowIdx].label}')
                if self.optFunInfo[rowIdx].note is not None:
                    out.write(f' ({self.optFunInfo[rowIdx].note})')
            out.write('\n')

    def normalize(self, rowIdx, colIdx):
        row = self.tableau[rowIdx]
        r = row[colIdx]
        if r == 1:
            return
        if r == 0:
            raise ZeroDivisionError
        for i, rate in enumerate(row):
            row[i] = div(rate, r)
            
    def clear(self, rowIdx, colIdx):
        self._cleared = None
        self.normalize(rowIdx, colIdx)
        for i, row in enumerate(self.tableau):
            if i == rowIdx: continue
            factor = row[colIdx]
            if factor == 0: continue
            for j, val in enumerate(self.tableau[rowIdx]):
                row[j] -= factor*val
        self.basic[rowIdx] = self.cols[colIdx]
        #assert(self.tableau[rowIdx][colIdx] == 1)

    def fixClearedCol(self, rowIdx, colIdx, rowIdxToFix):
        #assert(self.tableau[rowIdx][colIdx] == 1)
        row = self.tableau[rowIdxToFix]
        factor = row[colIdx]
        for j, val in enumerate(self.tableau[rowIdx]):
            row[j] -= factor*val
        
    def clearCol(self, colIdx):
        rowIdx = self.bestRow(colIdx)
        if rowIdx is not None:
            self.clear(rowIdx, colIdx)
        return rowIdx

    def zeroCol(self, colIdx):
        for row in self.tableau:
            row[colIdx] = 0
        self.zeroed[colIdx] = True
        
    def solution(self):
        res = {var: 0 for var in self.cols if type(var) is Var}
        other = {}
        for rowIdx, var in enumerate(self.basic):
            colIdx = self.colsByVar[var]
            assert self.tableau[rowIdx][colIdx] == 1
            sol = self.tableau[rowIdx][-1]
            if type(var) is Var:
                res[var] = sol
            elif type(var) is SlackVar and sol < 0:
                other[var] = sol
            elif isinstance(var, ArtificialVar):
                other[var] = sol
        return Solution(res,other)
    
    def allSolutions(self):
        assert self.optFunBegin == len(self.tableau)
        prev = {}
        sols = []
        # FIXME: Is this guaranteed to find all solutions?
        while True:
            encoding = self.encoding()
            if encoding in prev:
                colIdx = prev[encoding]
            else:
                sols.append(self.solution())
                colIdx = -1
            try:
                colIdx = self.cleared().index(-1, colIdx + 1)
            except ValueError:
                break
            prev[encoding] = colIdx
            self.clearCol(colIdx)
        return sols
            
    def ratios(self, colIdx):
        res = {}
        for i, row in enumerate(self.tableau[:self.optFunBegin]):
            #if row[-1] >= 0 and row[colIdx] > 0:
            if row[colIdx] > 0:
                r = div(row[-1], row[colIdx])
                if r >= 0:
                    res[i] = r
        return res
    
    def bestRow(self, colIdx, start = 0):
        try:
            row, ratio = min(self.ratios(colIdx).items(), key=operator.itemgetter(1))
        except ValueError:
            return None
        else:
            return row
        
    def cleared(self):
        if self._cleared:
            return self._cleared
        """-2: all zero, -1: not cleared, >= 0: colIdx"""
        cleared = [-2]*len(self.cols)
        for rowIdx, var in enumerate(self.basic):
            colIdx = self.colsByVar[var]
            cleared[colIdx] = rowIdx
        for colIdx, c in enumerate(cleared):
            if c == -2: 
                for rowIdx in range(len(self.basic)):
                    if self.tableau[rowIdx][colIdx] != 0:
                        cleared[colIdx] = -1
        self._cleared = cleared
        return cleared
    
    def encoding(self):
        cleared = self.cleared()
        l = len(cleared) - 1
        return sum(0 if c < 0 else 1<<(l-i) for i,c in enumerate(cleared))
    
    #def copy(self): # broken, needs updating
    #    other = object.__new__(type(self))
    #    other.cols = self.cols
    #    other.colsByVar = self.colsByVar
    #    other.tableau = [row.copy() for row in self.tableau]
    #    other.optFunBegin = self.optFunBegin
    #    other.cleared = self.cleared.copy()
    #    return other
    
    def addObjective(self, terms, note = None, label = 'max'):
        if label not in ('max', 'aux'):
            raise ValueError("label must be 'max' or 'aux'")
        row = [0]*(len(self.cols) + 1)
        cleared = self.cleared()
        toClear = []
        for var, rate in terms.items():
            if type(var) is int:
                colIdx = var
            else:
                try:
                    colIdx = self.colsByVar[var]
                except KeyError:
                    continue
            if self.zeroed[colIdx]:
                continue
            row[colIdx] = -rate
            if cleared[colIdx] >= 0:
                toClear.append(colIdx)
        self.tableau.append(row)
        optFun = _OptFun(label, terms, note)
        self.optFunInfo.append(optFun)
        self.allOptFuns.append(optFun)
        for colIdx in toClear:
            self.fixClearedCol(cleared[colIdx], colIdx, -1)

    def max(self, rowIdx):
        if rowIdx < self.optFunBegin:
            raise ValueError(f'row {rowIdx} not an objective function')
        opt = self.tableau[rowIdx]
        res = SolveRes.NOOP
        while True:
            col, val = None, 0
            for j in range(0, len(opt) - 1):
                if opt[j] < val:
                    col = j
                    val = opt[j]
            if val == 0:
                break
            res |= SolveRes.OPTIMAL
            row = self.clearCol(col)
            if row is None:
                # The solution is likely unbounded, set the var to 0 to find a
                # partial solution
                self.zeroCol(col)
                self._unboundedVars.append(self.cols[col])
                res |= SolveRes.UNBOUNDED
        return res

    def solve(self, rowIdx = None, *, zero = False):
        if rowIdx is None:
            rowIdx = self.optFunBegin
            if rowIdx == len(self.tableau):
                return SolveRes.NOOP
        if rowIdx < self.optFunBegin:
            raise ValueError(f'row {rowIdx} not an objective function')
        
        opt = self.tableau[rowIdx]
        res = SolveRes.OPTIMAL
        if sum(v != 0 for v in opt) == 1:
            res = SolveRes.UNIQUE
        res |= self.max(rowIdx)

        unique = None
        # first determine the maxium value for each aux function
        i = rowIdx + 1
        while i < len(self.optFunInfo) and self.optFunInfo[i].label == 'aux':
            res0 = self.max(i)
            assert(res0.ok())
            self.optFunInfo[i].max = self.tableau[i][-1]
            i += 1
        # now fix the result
        res0 = self.max(rowIdx)
        assert(res0.ok())
        # now check that each aux function is at it's max
        i = rowIdx + 1
        while i < len(self.optFunInfo) and self.optFunInfo[i].label == 'aux':
            if self.optFunInfo[i].max > self.tableau[i][-1]:
                unique = False
                self.optFunInfo[i].optimal = False
            elif unique is None:
                unique = True
            i += 1
        if unique is True:
            res |= SolveRes.UNIQUE
        elif unique is False:
            res |= SolveRes.OK
            
        if zero:
            toClear = []
            for colIdx in range(0, len(opt) - 1):
                if type(self.cols[colIdx]) is Var:
                    if opt[colIdx] != 0:
                        self.zeroCol(colIdx)
                else:
                    if self.zeroed[colIdx] or opt[colIdx] != 0:
                        self.zeroed[colIdx] = False
                        toClear.append(colIdx)
            delStop = rowIdx + 1
            while delStop < len(self.optFunInfo) and self.optFunInfo[delStop].label == 'aux':
                delStop += 1
            del self.tableau[rowIdx:delStop]
            del self.optFunInfo[rowIdx:delStop]
            if toClear:
                _multiDel(self.cols, toClear)
                for row in self.tableau:
                    _multiDel(row, toClear)
            self.colsByVar = {var: idx for idx, var in enumerate(self.cols)}
        return res

    def addPendingObjective(self):
        lst = self.pendingObjectives.pop(0)
        itr = iter(lst)
        self.addObjective(*next(itr))
        for aux in itr:
            self.addObjective(*aux,'aux')
        
    def solveAll(self):
        res = SolveRes.NOOP
        while len(self.tableau) > self.optFunBegin:
            res |= self.solve(zero = True)
        while self.pendingObjectives:
            self.addPendingObjective()
            res |= self.solve(zero = True)
        if any(c == -1 for c in self.cleared()):
            res |= SolveRes.MULTI
        return res

def _multiDel(lst, idxsToDel):
    try:
        nextToDel = idxsToDel[0]
    except IndexError:
        return
    origIdx, newIdx = nextToDel, nextToDel
    i = 1
    origLen = len(lst)
    while origIdx < origLen:
        if origIdx == nextToDel:
            if i < len(idxsToDel):
                nextToDel = idxsToDel[i]
                i += 1
            else:
                nextToDel = -1
            origIdx += 1
        else:
            lst[newIdx] = lst[origIdx]
            origIdx += 1
            newIdx += 1
    del lst[newIdx:]                

class Solution(NamedTuple):
    solution: dict
    other: dict
    def print(self, out = None):
        if out is None:
            out = sys.stdout
        out.write('Solution:\n')
        for var, rate in self.solution.items():
            out.write(f'  {var} = {rate:.9g}\n')
        out.write('Other:\n')
        for var, rate in self.other.items():
            out.write(f'  {var} = {rate:.9g}\n')

def walk(t): 
    # note: likely broken
    prev = set()
    solutions = set()
    colLen, rowLen = len(t.cols), len(t.tableau)
    cnt = 0
    def step():
        nonlocal t, cnt
        cnt += 1
        prev.add(t.encoding())
        #print(f'{t.encoding():0{colLen}b} {cnt}')
        sol = t.solution()
        if not sol[1] and all(rate >= 0 for rate in sol[0].values()):
        #if all(rate >= 0 and rate <= 1 for rate in sol[0].values()):
            print(*(f'{rate:g}' for rate in sol[0].values()))
            solutions.add(tuple(sol[0].values()))
        for colIdx in range(colLen):
            if t.cleared[colIdx] > 0:
                continue
            for rowIdx in range(rowLen):
                if t.tableau[rowIdx][colIdx] != 0:
                    t1 = t.copy()
                    t1.clear(rowIdx, colIdx)
                    if t1.encoding() not in prev:
                        t = t1
                        return True
        return False
    while step():
        pass
    return solutions





from __future__ import annotations
import draftsman.data.items as d_items
import draftsman.data.modules as d_modules
import draftsman.data.recipes as d_recipes
from . import itm,rcp,rcpinst,data,machines as mch
from .fracs import frac
from .core import *
from .data import *
from .data import CraftingHint, craftingHints
from .machines import *
from ._helper import toPythonName

def doit():
    addItem(Electricity, 'electricity', 'zzz')

    for (k,v) in d_modules.raw.items():
        pythonName = toPythonName(k)
        def get(what):
            return frac(v['effect'].get(what, {'bonus': 0})['bonus'], float_conv_method = 'round')
        e = Effect(speed = get('speed'),
                   productivity = get('productivity'),
                   consumption = get('consumption'),
                   pollution = get('pollution'))
        limitation = v.get('limitation', None)
        if limitation is not None:
            limitation = set(limitation)
            if 'rocket-part' in limitation:
                limitation.add('space-science-pack')
        item = Module(k, v['order'], v['stack_size'], e, limitation)
        setattr(itm, pythonName, item)
        itm.byName[k]=item
    rcpinst.byName[None] = {}
    rcpinst.byName[Mode.NORMAL] = {}
    rcpinst.byName[Mode.EXPENSIVE] = {}
    for (k,v) in d_recipes.raw.items():
        category = v.get('category','crafting')
        madeIn = categoryToMachine[category]
        def toRecipeComponent(d):
            if type(d) is dict:
                num = d['amount']*d.get('probability',1)
                if type(num) is float:
                    num = frac(num, float_conv_method = 'round')
                return RecipeComponent(item=lookupItem(d['name']), num = num)
            else:
                return RecipeComponent(item=lookupItem(d[0]),num=d[1])
        def toRecipe(d):
            inputs = tuple(toRecipeComponent(rc) for rc in d['ingredients'])
            try:
                results = d['results']
            except KeyError:
                results = [(d['result'], d.get('result_count', 1))]
            outputs = tuple(toRecipeComponent(rc) for rc in results)
            time = frac(d.get('energy_required', 0.5), float_conv_method = 'round')
            order = v.get('order',None)
            if order is None and len(outputs) == 1:
                order = outputs[0].item.order
            if order is None and 'main_product' in d:
                order = itm.byName[d['main_product']].order
            assert(order is not None)
            return Recipe(v['name'],madeIn,inputs,outputs,time,order)
        try:
            normal_recipe = toRecipe(v['normal'])
            expensive_recipe = toRecipe(v['expensive'])
        except KeyError:
            recipe = toRecipe(v)
            normal_recipe = recipe
            expensive_recipe = recipe
        addRecipe(normal_recipe, expensive_recipe)

    rp = rcpinst.byName[None]['rocket-part']
    rocket_parts_inputs = tuple(RecipeComponent(rc.num*100, rc.item) for rc in rp.inputs)
    rocket_parts_time = rp.time*100

    space_science_pack = RocketSilo.Recipe(
        name = 'space-science-pack',
        madeIn = RocketSilo,
        order = lookupItem('space-science-pack').order,
        inputs = rocket_parts_inputs,
        outputs = (RecipeComponent(num=1000,
                                   item=lookupItem('space-science-pack')),),
        time = rocket_parts_time,
        cargo = RecipeComponent(num=1,
                                item=itm.satellite))
    addRecipe(space_science_pack,space_science_pack)

    steam = Recipe(
        name = 'steam',
        madeIn = Boiler,
        inputs = (RecipeComponent(60, lookupItem('water')),),
        outputs = (RecipeComponent(60, lookupItem('steam')),),
        time = 1,
        order = '')
    addRecipe(steam,steam)

    researchHacks()

    for r in rcpinst.byName[Mode.NORMAL].values():
        for _, item in r.outputs:
            recipesThatMake.setdefault(item, []).append(r.name)
        for _, item in r.inputs:
            recipesThatUse.setdefault(item, []).append(r.name)
        if any(item == itm.empty_barrel for _, item in r.outputs) and len(r.outputs) > 1:
            craftingHints[r.name] = CraftingHint(priority = IGNORE)

    craftingHints['advanced-oil-processing'] = CraftingHint(also=['light-oil-cracking','heavy-oil-cracking'])
    craftingHints['coal-liquefaction']       = CraftingHint(also=['light-oil-cracking','heavy-oil-cracking'])

    craftingHints['light-oil-cracking'] = CraftingHint(priority=-1)
    craftingHints['heavy-oil-cracking'] = CraftingHint(priority=-1)

    craftingHints['solid-fuel-from-light-oil']     = CraftingHint(priority =0, also=['solid-fuel-from-petroleum-gas'])
    craftingHints['solid-fuel-from-petroleum-gas'] = CraftingHint(priority=-1)
    craftingHints['solid-fuel-from-heavy-oil']     = CraftingHint(priority=-2)

    craftingHints['nuclear-fuel-reprocessing']     = CraftingHint(boxPriority = 90)

    craftingHints['kovarex-enrichment-process']    = CraftingHint(priority = IGNORE)
    
    craftingHints['rocket-part'] = CraftingHint(priority = IGNORE)

    data.entityToMachine = {
        cls.name: cls for cls in (CraftingMachine, Boiler,
                                  AssemblingMachine1, AssemblingMachine2, AssemblingMachine3,
                                  ChemicalPlant, OilRefinery, Centrifuge,
                                  StoneFurnance, SteelFurnance, ElectricFurnace,
                                  RocketSilo,
                                  Beacon)
    }


def researchHacks():
    from .helper import sciencePacks
    addResearch('_production_research', 'zz0', sciencePacks - {itm.military_science_pack})
    addResearch('_military_research', 'zz1', sciencePacks - {itm.production_science_pack})
    addResearch('_combined_research', 'zz2', sciencePacks)

def addResearch(name, order, inputs):
    from .helper import FakeLab
    item = addItem(Research, name, order)
    recipe = Recipe(name = name,
                    madeIn = FakeLab,
                    inputs = (RecipeComponent(1, i) for i in sorted(inputs, key = lambda k: k.order)),
                    outputs = (RecipeComponent(1, item),),
                    time = 1,
                    order = order)
    addRecipe(recipe, recipe)

def addItem(cls, name, order):
    item = cls(name, order)
    setattr(itm, name, item)
    itm.byName[name]=item
    return item

def translateEnergyString(s):
    if s is None:
        return None
    elif s == '':
        return 0
    elif s.endswith('kJ'):
        return int(float(s[0:-2])*1_000)
    elif s.endswith('MJ'):
        return int(float(s[0:-2])*1_000_000)
    elif s.endswith('GJ'):
        return int(float(s[0:-2])*1_000_000_000)
    else:
        raise ValueError

def lookupItem(name):
    try:
        return itm.byName[name]
    except KeyError:
        pass
    pythonName = toPythonName(name)
    try:
        d = d_items.raw[name]
        item = Item(name,d['order'],d['stack_size'],fuelValue = translateEnergyString(d.get('fuel_value','')),fuelCategory = d.get('fuel_category',''))
    except KeyError:
        item = Ingredient(name,'z-'+name)
    setattr(itm, pythonName, item)
    itm.byName[name] = item
    return item

def addRecipe(recipe, expensive_recipe):
    name = recipe.name
    pythonName = toPythonName(name)
    rcp_ = object.__new__(Rcp)
    object.__setattr__(rcp_, 'name', name)
    setattr(rcp, pythonName, rcp_)
    rcp.byName[name] = rcp_
    if recipe is expensive_recipe:
        rcpinst.byName[None][name] = recipe
    setattr(rcpinst, pythonName, recipe)
    setattr(rcpinst.expensive, pythonName, expensive_recipe)
    rcpinst.byName[Mode.NORMAL][name] = recipe
    rcpinst.byName[Mode.EXPENSIVE][name] = expensive_recipe

categoryToMachine = {
    'crafting': mch.AssemblingMachine,
    "advanced-crafting": mch.AssemblingMachine,
    'crafting-with-fluid': mch.FluidCapableAssemblingMachine,
    'centrifuging': mch.Centrifuge,
    'chemistry': mch.ChemicalPlant,
    'oil-processing': mch.OilRefinery,
    'rocket-building': None, # a special case
    'smelting': mch.Furnace,
}



from dataclasses import dataclass

from .core import CraftingMachine, Recipe, RecipeComponent
from . import itm


__all__ = ('sciencePacks','FakeLab')

sciencePacks = {itm.automation_science_pack,
                itm.logistic_science_pack,
                itm.chemical_science_pack,
                itm.production_science_pack,
                itm.utility_science_pack,
                itm.space_science_pack,
                itm.military_science_pack}

@dataclass(init=False)
class FakeLab(CraftingMachine):
    pass




from __future__ import annotations
from dataclasses import dataclass
from collections import defaultdict
from collections.abc import Sequence

from .fracs import frac,div,diva
from .core import *
from .core import _MutableFlows
from . import itm, rcp

__all__ = ('Beacon', 'Boiler', 'AssemblingMachine1', 'AssemblingMachine2', 'AssemblingMachine3',
           'ChemicalPlant', 'OilRefinery', 'StoneFurnance', 'SteelFurnance', 'ElectricFurnace',
           'Centrifuge', 'RocketSilo')

@dataclass(init=False)
class _BurnerMixin:
    energyType = 'burner'
    fuel: Item = None

    def __init__(self, *args, fuel = None, **kws):
        from ._helper import asItem
        super().__init__(*args, **kws)
        fuel = asItem(fuel)
        self.fuel = fuel

    def _jsonObj(self, **kwargs):
        obj = super()._jsonObj(**kwargs)
        if type(obj) is int:
            return obj
        obj['fuel'] = self.fuel.name if self.fuel else None
        return obj

    def __setattr__(self, prop, val):
        if prop == 'fuel':
            if val is not None and val.fuelCategory != 'chemical':
                raise ValueError(f'invalid item for fuel: {val}')
        super().__setattr__(prop, val)

    def _calc_flows(self, throttle):
        flows = super()._calc_flows(throttle)
        fuelRate = div(self.energyUsage(throttle), self.fuel.fuelValue)
        flows.addFlow(self.fuel, rateIn = fuelRate, adjusted = self.throttle != 1)
        return flows

@dataclass(init=False)
class _ElectricMixin:
    energyType = 'electric'

    def _calc_flows(self, throttle):
        flows = super()._calc_flows(throttle)
        flows.addFlow(itm.electricity, rateIn = div(self.energyUsage(throttle), 1_000_000))
        return flows

@dataclass(init=False)
class _ModulesMixin:
    modules: tuple[Module, ...]
    beacons: list[Beacon]

    def __init__(self, *args, modules = None, beacons = None, **kws):
        super().__init__(*args, **kws)
        self.modules = () if modules is None else modules
        self.beacons = [] if beacons is None else beacons

    def _jsonObj(self, **kwargs):
        obj = super()._jsonObj(**kwargs)
        if type(obj) is int:
            return obj
        if self.modules:
            obj['modules'] = [m._jsonObj() for m in self.modules]
        if self.beacons:
            obj['beacons'] = [b._jsonObj(**kwargs) for b in self.beacons]
        return obj

    def _checkModules(self, recipe, modules):
        invalid = set()
        for m in modules:
            if m.limitation is not None and recipe.name not in m.limitation:
                invalid.add(m)
        if invalid:
            raise InvalidModulesError(invalid)

    def __setattr__(self, prop, val):
        if prop == 'recipe':
            modules = getattr(self, 'modules', None)
            if modules:
                self._checkModules(val, modules)
        elif prop == 'modules' and self.recipe is not None:
            modules = []
            for v in val:
                if isinstance(v, Sequence):
                    for m in v:
                        modules.append(m)
                else:
                    modules.append(v)
            self._checkModules(self.recipe, modules)
            val = tuple(modules)
        elif prop == 'beacons' and isinstance(val, Mul):
            val = val.num*[val.machine]
        elif prop == 'beacons':
            beacons = []
            for v in val:
                if isinstance(v, Mul):
                    beacons += v.num*[v.machine]
                else:
                    beacons.append(v)
            val = beacons
        return super().__setattr__(prop, val)

    def _modulesStr(self):
        def fmt_w_num(num, obj):
            if num == 1:
                return str(obj)
            else:
                return f'{num:g} {obj}'
        modules = defaultdict(lambda: 0)
        for m in self.modules:
            modules[m] += 1
        modulesStr = ', '.join(fmt_w_num(num, m) for m, num in modules.items())
        beacons = defaultdict(lambda: 0)
        for b in self.beacons:
            beacons[b] += 1
        beaconsStr = ', '.join(fmt_w_num(num, b) for b, num in beacons.items())
        if modulesStr and beaconsStr:
            return f'{modulesStr}; {beaconsStr}'
        elif modulesStr:
            return modulesStr
        elif beaconsStr:
            return beaconsStr
        else:
            return ''

    def bonus(self):
        if len(self.modules) == 1 and len(self.beacons) == 0:
            # special case in case we have a FakeModule
            return Bonus(self.modules[0].effect)
        return Bonus(sum([m.effect for m in self.modules],Effect()) + sum([b.effect() for b in self.beacons],Effect()))

@dataclass(init=False)
class Beacon(_ElectricMixin,Machine):
    name = 'beacon'
    width = 3
    height = 3
    baseEnergyUsage = 0
    energyDrain = 480_000

    modules: list[Module]

    def __hash__(self):
        return hash((self.__class__, *self.modules))

    def _jsonObj(self, **kwargs):
        obj = super()._jsonObj(**kwargs)
        if type(obj) is int:
            return obj
        obj['modules'] = [m._jsonObj() for m in self.modules]
        return obj

    def __init__(self, modules = None, **kws):
        super().__init__(**kws)
        self.modules = [] if modules is None else modules
    
    def effect(self):
        return sum([m.effect for m in self.modules],Effect()) / 2

    def _modulesStr(self):
        modules = defaultdict(lambda: 0)
        for m in self.modules:
            modules[m] += 1
        return ', '.join(str(m) if num == 1 else f'{num:g} {m}' for m, num in modules.items())

@dataclass(init=False)
class Boiler(_BurnerMixin,CraftingMachine):
    name = "boiler"
    baseEnergyUsage = 1_800_000
    energyDrain = 0
    pollution = 30

    def __init__(self, **kws):
        super().__init__(**kws)
        self.recipe = rcp.steam

@dataclass(init=False)
class AssemblingMachine(_ElectricMixin,CraftingMachine):
    name = "assembling-machine"
    width = 3
    height = 3
    pass
AssemblingMachine.abstract = AssemblingMachine

@dataclass(init=False)
class FluidCapableAssemblingMachine(AssemblingMachine):
    pass
FluidCapableAssemblingMachine.abstract = FluidCapableAssemblingMachine

@dataclass(init=False)
class AssemblingMachine1(AssemblingMachine):
    name = "assembling-machine-1"
    craftingSpeed = frac(1,2)
    baseEnergyUsage = 75_000
    energyDrain = 2_500
    pollution = 4

@dataclass(init=False)
class AssemblingMachine2(_ModulesMixin,FluidCapableAssemblingMachine):
    name = "assembling-machine-2"
    craftingSpeed = frac(3,4)
    baseEnergyUsage = 150_000
    energyDrain = 5_000
    pollution = 3

@dataclass(init=False)
class AssemblingMachine3(_ModulesMixin,FluidCapableAssemblingMachine):
    name = "assembling-machine-3"
    craftingSpeed = frac(5,4)
    baseEnergyUsage = 375_000
    energyDrain = 12_500
    pollution = 2

@dataclass(init=False)
class ChemicalPlant(_ModulesMixin,_ElectricMixin,CraftingMachine):
    name = "chemical-plant"
    width = 3
    height = 3
    craftingSpeed = 1
    baseEnergyUsage = 210_000
    energyDrain = 7_000
    pollution = 4

@dataclass(init=False)
class OilRefinery(_ModulesMixin,_ElectricMixin,CraftingMachine):
    name = "oil-refinery"
    width = 5
    height = 5
    craftingSpeed = 1
    baseEnergyUsage = 420_000
    energyDrain = 14_000
    pollution = 6

@dataclass(init=False)
class Centrifuge(_ModulesMixin,_ElectricMixin,CraftingMachine):
    name = "centrifuge"
    width = 3
    height = 3
    craftingSpeed = 1
    baseEnergyUsage = 350_000
    energyDrain = 11_600
    pollution = 4

@dataclass(init=False)
class Furnace(CraftingMachine):
    """Generic furnace.  Does not map to a factory entity."""
    pass
Furnace.abstract = Furnace

@dataclass(init=False)
class StoneFurnance(_BurnerMixin,Furnace):
    name = "stone-furnace"
    width = 2
    height = 2
    craftingSpeed = 1
    baseEnergyUsage = 90_000
    energyDrain = 0
    pollution = 2

@dataclass(init=False)
class SteelFurnance(_BurnerMixin,Furnace):
    name = "steel-furnace"
    width = 2
    height = 2
    craftingSpeed = 2
    baseEnergyUsage = 90_000
    energyDrain = 0
    pollution = 4

@dataclass(init=False)
class ElectricFurnace(_ModulesMixin,_ElectricMixin,Furnace):
    name = "electric-furnace"
    width = 3
    height = 3
    craftingSpeed = 2
    baseEnergyUsage = 180_000
    energyDrain = 6_000
    pollution = 1

@dataclass(init=False)
class RocketSilo(_ModulesMixin,_ElectricMixin,CraftingMachine):
    class Recipe(Recipe):
        __slots__ = ('cargo')
        def __init__(self, name, madeIn, inputs, outputs, time, order, cargo):
            super().__init__(name, madeIn, inputs, outputs, time, order)
            object.__setattr__(self, 'cargo', cargo)
        
    name = "rocket-silo"
    width = 9
    height = 9
    craftingSpeed = 1
    baseEnergyUsage = 4_000_000
    energyDrain = 0
    delay = frac(2420 + 13, 60)

    def _calc_flows(self, throttle):
        recipe = self.recipe
        cargo = recipe.cargo
        b = self.bonus()
        time = diva(recipe.time, self.craftingSpeed, 1 + b.speed, 1 + b.productivity) + self.delay
        flows = _MutableFlows()
        for rc in self.recipe.inputs:
            rate = div(rc.num, time * (1 + b.productivity))
            flows.addFlow(rc.item, rateIn = throttle*rate, adjusted = throttle != 1)
        for rc in self.recipe.outputs:
            rate = div(rc.num, time)
            flows.addFlow(rc.item, rateOut = throttle*rate, adjusted = throttle != 1)
        flows.addFlow(cargo.item,
                      rateIn = div(throttle*cargo.num, time),
                      adjusted = throttle != 1)
        return flows





from typing import Any
from collections import deque
from collections.abc import Mapping
import math
from copy import copy
from dataclasses import dataclass,field
import operator

from . import itm,rcp,rcpinst,config
from .fracs import frac, div
from .core import *
from .box import *
from .solver import LinearEqSystem,Solver,SolveRes
from .data import *
from .data import craftingHints

__all__ = ('box', 'unboundedBox', 'produce', 'merge', 'union',
           'ProduceResult', 'MultipleChoiceError', 'SolveFailedError', 'NonUniqueSolutionError')

def _box(inner, minSolveRes, cls, **kwargs):
    if minSolveRes is None:
        minSolveRes = SolveRes.MULTI
    b = cls(inner, **kwargs)
    res = b.solve()
    if res > minSolveRes:
        if not res.ok():
            raise SolveFailedError(_BoxFailedInfo(b, res))
        raise NonUniqueSolutionError(_BoxFailedInfo(b, res))
    return b

def box(inner, *, minSolveRes = None, **kwargs):
    """Create a `Box` and then solve it.

    Will raise `NonUniqueSolutionError` if the solver returned a result larger
    than *minSolveRes* (default `SolveRes.MULTI`) or `SolveFailedError` if the
    solve failed.

    All other parameters are passed to `Box`.

    """
    return _box(inner, minSolveRes, Box, **kwargs)

def unboundedBox(inner, *, minSolveRes = None, **kwargs):
    """Create an `UnboundedBox` and then solve it.
    
    See `box`.

    """
    return _box(inner, minSolveRes, UnboundedBox, **kwargs)

def blackBox(inner, *, minSolveRes = None, **kwargs):
    """Create a `BlackBox` and then solve it.

    See `box`.
    """
    b = _box(inner, minSolveRes, Box, **kwargs)
    return BlackBox(b)

@dataclass
class _BoxFailedInfo:
    factory: Box = None
    solveRes: SolveRes = SolveRes.UNSOLVED
    

def produce(outputs, using = (), *, stopAt = (), fuel = None,
            name = None,
            recursive = True, roundUp = False, minSolveRes = SolveRes.MULTI, solve = True):
    """Create a factory to produce outputs.

    It works in two stages the first is select appropriate recipes and
    machines to produce the given outputs and the second is to determine the
    nunber of machines needed using the same solver used by the solve method
    in Box.

    The paramaters are:

    *outputs*
      sequence or mapping of outputs, rates are per second and can be
      `None` if inputs are well specified.

    *using*
      an optional mixed type sequence used to guide the selection of
      what recipes to use and when to stop.
 
      If an item is given it will try and select recipes that use that item
      and stop once it is used.  If paired with a rate (using the '@'
      operator) then will use that item at the given rate.

      If a recipe is given the solver will attempt to use that recipe if
      possible and keep going.  If more than one recipe is specified for a
      given output both will be used and the solver step will attemt to
      determine the best combination.

      If a machine instance is given that machine configuration will be used
      when applicable.  If a machine instance recipe is not None that
      machine configuration will only be used for the given recipe.  If a
      matching machine is not found, `produce` will use
      `config.machinePrefs`.

    *fuel*
      item to use as fuel when applicable, if None will use
      `config.defaultFuel`

    *recursive*
      |nbsp|

    *roundUp*
      Round up multiple of machines to integer values and adjust the
      throttle to compensate.

    *minSolveRes*
      The minimum acceptable solver result.  Defaults to
      `SolveRes.MULTI`.

    The result is a dataclass with the following fields:

    factory
      the resulting factory as an instance of Box

    solveRes
      the result of the second stage solver

    extraOutputs
      list of auxiliary outputs; if these outputs are not consumed
      somehow than the entiry factory is likely to grind to a halt

    extraInputs
      list of inputs used not specified in `using` if no inputs are
      specified this list will be empty

    unusedInputs
      list of unused inputs

    May also throw the following errors (all subclass ValueError)

    * `MultipleChoiceError`
    * `SolveFailedError`
    * `NonUniqueSolutionError`

    """
    outputs = Box.Outputs(outputs)
    inputs = Box.Inputs()
    if fuel is None:
        fuel = config.defaultFuel.get()
    mode = config.mode.get()
    recipePrefs = {}
    machinePrefs = []
    for v in using.items() if isinstance(using, Mapping) else using:
        if isinstance(v, Ingredient):
            inputs[v] = None
        elif isinstance(v, Rcp):
            recipePrefs[v.name] = 1000
        elif isinstance(v, Machine):
            machinePrefs.append(v)
        elif type(v) is tuple:
            item, rate = v
            if not isinstance(item, Ingredient):
                raise ValueError
            inputs[item] = rate
        else:
            raise ValueError(f"don't know how to use {v}")
    stopAt = set(stopAt)
    stopAt |= inputs.keys()
    machinePrefs += config.machinePrefs.get([])
    l = list(inputs.keys())
    origLen = len(l)
    i = 0
    while i < len(l):
        priority = 1000 if i < origLen else 800
        item = l[i]
        i += 1
        rs = recipesThatUse.get(item, ())
        if len(rs) == 1:
            r = rs[0]
            recipePrefs[r] = max(recipePrefs.get(r, priority), priority)
            recipe = rcpinst.byName[mode][r]
            m = recipe(machinePrefs = machinePrefs, fuel = fuel)
            for flow in m.flows():
                if flow.rate() > 0 and flow.item not in l:
                    l.append(flow.item)
        else:
            for r in rs:
                try:
                    adj = craftingHints[r].priority
                    if adj <= IGNORE:
                        continue
                except KeyError:
                    adj = 0
                p = priority - 100 + adj
                recipePrefs[r] = max(recipePrefs.get(r, p), p) 
    machines = {}
    toResolve = deque(outputs.keys())
    while toResolve:
        item = toResolve.popleft()
        try:
            recipes = recipesThatMake[item]
        except KeyError:
            continue
        maxPriority = None
        bestRecipes = []
        for r in recipes:
            try:
                priority = recipePrefs[r]
            except KeyError:
                try:
                    priority = craftingHints[r].priority
                except KeyError:
                    priority = 0
            if maxPriority is None or priority > maxPriority:
                maxPriority = priority
                bestRecipes = [r]
            elif priority == maxPriority:
                bestRecipes.append(r)
        else:
            if maxPriority <= IGNORE:
                continue
        if len(bestRecipes) == 1 or (len(bestRecipes) > 1 and maxPriority >= 1000):
            recipes = bestRecipes
        else:
            raise MultipleChoiceError('multiple ways to produce {}: {}'.format(item, ' '.join(bestRecipes)))
        recipes = [r for r in recipes if r not in machines]
        if not recipes:
            continue
        try:
            recipes += craftingHints[recipes[0]].also
        except KeyError:
            pass
        for r in recipes:
            if r in machines:
                continue
            recipe = rcpinst.byName[mode][r]
            m = recipe(machinePrefs = machinePrefs, fuel = fuel)
            machines[r] = m
            if recursive:
                for flow in m.flows():
                    if flow.rate() < 0 and flow.item not in stopAt:
                        toResolve.append(flow.item)
    res = ProduceResult()
    if not machines:
        return res
    boxOutputs = dict(outputs)
    boxPriorities = {}
    for recipe, m in machines.items():
        try:
            if craftingHints[recipe].boxPriority != 0:
                boxPriorities[m.recipe] = craftingHints[recipe].boxPriority
        except KeyError:
            pass
        m_outputs = m.outputs
        if len(m_outputs) > 1:
            for item in m_outputs.keys():
                if item not in boxOutputs:
                    boxOutputs[item] = None
    b = res.factory = UnboundedBox(name = name,
                                   inner = Group([Mul(1, m) for m in machines.values()]),
                                   outputs = boxOutputs, inputTouchups = inputs,
                                   priorities = boxPriorities)
    someOutputRatesSpecified = any(r != None for r in outputs.values())
    for item in b.outputs.keys():
        if item not in outputs:
            b.priorities[item] = IGNORE
        elif someOutputRatesSpecified and outputs[item] is None:
            b.priorities[item] = IGNORE
    if inputs:
        for item in b.inputs.keys():
            if item not in inputs: 
                res.extraInputs.append(item)
    if not solve or (all(rate is None for rate in outputs.values()) and all(rate is None for rate in inputs.values())):
        return res
    res.solveRes = b.solve()
    b.finalize(roundUp = roundUp, _res = res, recursive = False)
    if res.solveRes > minSolveRes:
        if not res.solveRes.ok():
            raise SolveFailedError(res)
        raise NonUniqueSolutionError(res)
    return res

@dataclass
class ProduceResult:
    factory: Box = None
    solveRes: SolveRes = SolveRes.UNSOLVED
    extraOutputs: list = field(default_factory=list)
    extraInputs: list = field(default_factory=list)
    unusedInputs: list = field(default_factory=list) 

class MultipleChoiceError(ValueError):
    pass

class SolveFailedError(ValueError):
    def __init__(self, res):
        self.res = res
        super().__init__(res.solveRes)

class NonUniqueSolutionError(ValueError):
    def __init__(self, res):
        self.res = res
        super().__init__(res.solveRes)

def merge(*args, mergeFun = operator.add):
    if len(args) < 2:
        raise ValueError
    res = _merge(args[0], args[1], mergeFun)
    for i in range(2, len(args)):
        res = _merge(res, args[i], mergeFun)
    views = [Box(res, name = b.name,
                 outputs = b.outputs, inputs = b.inputs,
                 constraints = b.constraints, priorities = b.priorities) for b in args]
    return (res,*views)

def union(*args):
    return merge(*args, mergeFun = max)
    
def _merge(b1, b2, mergeFun):
    combined = {}
    for m in b1.inner:
        num = m.num
        m = m.machine
        if isinstance(m, Box):
            combined[id(m)] = (num, m)
        else:
            combined[m.recipe] = (num, m)
    for m in b2.inner:
        num = m.num
        m = m.machine
        if isinstance(m, Box):
            existing = combined.get(id(m), (0, None))
            combined[id(m)] = (mergeFun(existing[0], num), m)
        else:
            existing = combined.get(m.recipe, (0, None))
            if existing[1]:
                assert type(m) is type(existing[1])
                assert m.bonus() == existing[1].bonus()
            combined[m.recipe] = (mergeFun(existing[0], num), m)

    outputs = dict()
    outputs.update((item, None) for item in b1.outputs.keys())
    outputs.update((item, None) for item in b2.outputs.keys())
    for item in outputs.keys():
        o1 = b1.outputs.get(item, None)
        o2 = b2.outputs.get(item, None)
        if o1 == o2:
            outputs[item] = o1
        
    inputs = dict()
    inputs.update((item, None) for item in b1.inputs.keys())
    inputs.update((item, None) for item in b2.inputs.keys())
    for item in inputs.keys():
        o1 = b1.inputs.get(item, None)
        o2 = b2.inputs.get(item, None)
        if o1 == o2:
            inputs[item] = o1
    return Box(Group(num*copy(m) for num,m in combined.values()), outputs=outputs, inputs=inputs)



import enum

class OrdEnum(enum.Enum):
    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented
    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented
    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented
    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented
    def __bool__(self):
        return bool(self.value)
    def __and__(self, other):
        return type(self)(min(self.value, other.value))
    def __or__(self, other):
        return type(self)(max(self.value, other.value))



__all__ = ['itm', 'rcp', 'rcpinst', 'config']

from .fracs import frac, div, Frac
__all__ += ['frac', 'div', 'Frac']

from .core import *
__all__ += core.__all__

from .machines import *
__all__ += machines.__all__

from .units import *
__all__ += units.__all__

from .data import *
__all__ += data.__all__

from . import _import_raw
_import_raw.doit()

from .presets import *
__all__ += presets.__all__

from . import box as _box
from .box import *
__all__ += _box.__all__

from .solver import SolveRes
__all__ += ['SolveRes']

from .blueprint import *
__all__ += blueprint.__all__

from .jsonconv import *
__all__ += jsonconv.__all__

from . import produce as _produce
from .produce import *
__all__ += _produce.__all__

from .helper import *
__all__ += helper.__all__

## extra symbols not export by default
from .ordenum import OrdEnum
from .core import Uniq, Immutable
_extraSymbols = ['Uniq', 'Immutable', 'OrdEnum']
